# -*- coding: utf-8 -*-
#
# Copyright 2017-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku workflow commands."""


import uuid
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional

from git import Actor

from renku.core import errors
from renku.core.commands.format.workflow import WORKFLOW_FORMATS
from renku.core.commands.view_model import plan_view
from renku.core.commands.view_model.composite_plan import CompositePlanViewModel
from renku.core.management.command_builder import inject
from renku.core.management.command_builder.command import Command
from renku.core.management.interface.activity_gateway import IActivityGateway
from renku.core.management.interface.client_dispatcher import IClientDispatcher
from renku.core.management.interface.plan_gateway import IPlanGateway
from renku.core.management.interface.project_gateway import IProjectGateway
from renku.core.management.workflow.concrete_execution_graph import ExecutionGraph
from renku.core.management.workflow.plan_factory import delete_indirect_files_list
from renku.core.management.workflow.value_resolution import CompositePlanValueResolver, ValueResolver
from renku.core.models.provenance.activity import Activity, ActivityCollection
from renku.core.models.workflow.composite_plan import CompositePlan
from renku.core.models.workflow.plan import AbstractPlan, Plan
from renku.core.plugins.provider import execute
from renku.core.utils import communication
from renku.core.utils.datetime8601 import local_now
from renku.core.utils.git import add_to_git
from renku.core.utils.os import get_relative_paths
from renku.version import __version__, version_url


def _ref(name):
    """Return workflow reference name."""
    return "workflows/{0}".format(name)


def _deref(ref):
    """Remove workflows prefix."""
    assert ref.startswith("workflows/")
    return ref[len("workflows/") :]


def _safe_read_yaml(file: str) -> Dict[str, Any]:
    try:
        from renku.core.models import jsonld as jsonld

        return jsonld.read_yaml(file)
    except Exception as e:
        raise errors.ParameterError(e)


@inject.autoparams()
def _find_workflow(name_or_id: str, plan_gateway: IPlanGateway) -> AbstractPlan:
    workflow = plan_gateway.get_by_id(name_or_id) or plan_gateway.get_by_name(name_or_id)

    if not workflow:
        raise errors.ParameterError(f'The specified workflow "{name_or_id}" cannot be found.')
    return workflow


@inject.autoparams()
def _list_workflows(plan_gateway: IPlanGateway, format: str, columns: List[str]):
    """List or manage workflows with subcommands."""
    workflows = plan_gateway.get_newest_plans_by_names()

    if format not in WORKFLOW_FORMATS:
        raise errors.UsageError(f'Provided format "{format}" is not supported ({", ".join(WORKFLOW_FORMATS.keys())})"')

    if format == "json-ld":
        return WORKFLOW_FORMATS[format](list(workflows.values()), columns=columns)

    return WORKFLOW_FORMATS[format](list(map(lambda x: plan_view(x), workflows.values())), columns=columns)


def list_workflows_command():
    """Command to list or manage workflows with subcommands."""
    return Command().command(_list_workflows).require_migration().with_database(write=False)


@inject.autoparams()
def _remove_workflow(name, force, plan_gateway: IPlanGateway):
    """Remove the remote named <name>."""
    workflows = plan_gateway.get_newest_plans_by_names()
    plan = None
    if name.startswith("/plans/"):
        plan = next(filter(lambda x: x.id == name, workflows.values()), None)
    if not plan and name not in workflows:
        raise errors.ParameterError(f'The specified workflow is "{name}" is not an active workflow.')

    if not force:
        prompt_text = f'You are about to remove the following workflow "{name}".' + "\n" + "\nDo you wish to continue?"
        communication.confirm(prompt_text, abort=True, warning=True)

    plan = plan or workflows[name]
    plan._v_immutable = False
    plan.invalidated_at = datetime.utcnow()
    plan.freeze()


def remove_workflow_command():
    """Command that removes the workflow named <name>."""
    return Command().command(_remove_workflow).require_clean().with_database(write=True).with_commit()


def _show_workflow(name_or_id: str):
    """Show the details of a workflow."""
    workflow = _find_workflow(name_or_id)
    return plan_view(workflow)


def show_workflow_command():
    """Command that the details of a workflow."""
    return Command().command(_show_workflow).require_migration().with_database(write=False)


@inject.autoparams()
def _group_workflow(
    name: str,
    description: str,
    mappings: List[str],
    defaults: List[str],
    links: List[str],
    param_descriptions: List[str],
    map_inputs: bool,
    map_outputs: bool,
    map_params: bool,
    link_all: bool,
    keywords: List[str],
    steps: List[str],
    plan_gateway: IPlanGateway,
    project_gateway: IProjectGateway,
) -> CompositePlan:
    """Group workflows into a CompositePlan."""

    if plan_gateway.get_by_name(name):
        raise errors.ParameterError(f"Duplicate workflow name: workflow '{name}' already exists.")

    child_workflows = []

    for workflow_name_or_id in steps:
        child_workflow = plan_gateway.get_by_id(workflow_name_or_id)

        if not child_workflow:
            child_workflow = plan_gateway.get_by_name(workflow_name_or_id)

        if not child_workflow:
            raise errors.ObjectNotFoundError(workflow_name_or_id)

        child_workflows.append(child_workflow)

    plan = CompositePlan(
        description=description,
        id=CompositePlan.generate_id(),
        keywords=keywords,
        name=name,
        plans=child_workflows,
        project_id=project_gateway.get_project().id,
    )

    if mappings:
        plan.set_mappings_from_strings(mappings)

    if defaults:
        plan.set_mapping_defaults(defaults)

    if links:
        plan.set_links_from_strings(links)
        graph = ExecutionGraph(plan)
        cycles = graph.cycles
        if cycles:
            cycles = [map(lambda x: x.name, cycle) for cycle in cycles]
            raise errors.GraphCycleError(cycles)

    if param_descriptions:
        plan.set_mapping_descriptions(param_descriptions)

    if map_inputs:
        plan.map_all_inputs()

    if map_outputs:
        plan.map_all_outputs()

    if map_params:
        plan.map_all_parameters()

    if link_all:
        # NOTE: propagate values to for linking to use
        rv = CompositePlanValueResolver(plan, None)
        plan = rv.apply()

        graph = ExecutionGraph(plan, virtual_links=True)

        cycles = graph.cycles
        if cycles:
            cycles = [map(lambda x: x.name, cycle) for cycle in cycles]
            raise errors.GraphCycleError(cycles)

        for virtual_link in graph.virtual_links:
            plan.add_link(virtual_link[0], [virtual_link[1]])

    plan_gateway.add(plan)

    return CompositePlanViewModel.from_composite_plan(plan)


def compose_workflow_command():
    """Command that creates a group of several workflows."""
    return (
        Command().command(_group_workflow).require_migration().require_clean().with_database(write=True).with_commit()
    )


@inject.autoparams()
def _edit_workflow(
    name,
    new_name: Optional[str],
    description: Optional[str],
    set_params: List[str],
    map_params: List[str],
    rename_params: List[str],
    describe_params: List[str],
    plan_gateway: IPlanGateway,
):
    """Edits a workflow details."""
    derived_from = _find_workflow(name)
    workflow = derived_from.derive()
    if new_name:
        workflow.name = new_name

    if description:
        workflow.description = description

    if isinstance(workflow, Plan):
        workflow.set_parameters_from_strings(set_params)

        def _kv_extract(kv_string):
            k, v = kv_string.split("=", maxsplit=1)
            v = v.strip(' "')
            return k, v

        for param_string in rename_params:
            name, new_name = _kv_extract(param_string)
            for param in workflow.inputs + workflow.outputs + workflow.parameters:
                if param.name == name:
                    param.name = new_name
                    break
            else:
                raise errors.ParameterNotFoundError(parameter=name, workflow=workflow.name)

        for description_string in describe_params:
            name, description = _kv_extract(description_string)
            for param in workflow.inputs + workflow.outputs + workflow.parameters:
                if param.name == name:
                    param.description = description
                    break
            else:
                raise errors.ParameterNotFoundError(parameter=name, workflow=workflow.name)
    elif isinstance(workflow, CompositePlan) and len(map_params):
        workflow.set_mappings_from_strings(map_params)

    plan_gateway.add(workflow)
    return plan_view(workflow)


def edit_workflow_command():
    """Command that edits the properties of a given workflow."""
    return Command().command(_edit_workflow).require_clean().with_database(write=True).with_commit()


@inject.autoparams()
def _export_workflow(
    name_or_id, client_dispatcher: IClientDispatcher, format: str, output: Optional[str], values: Optional[str]
):
    """Export a workflow to a given format."""
    client = client_dispatcher.current_client

    workflow = _find_workflow(name_or_id)
    if output:
        output = Path(output)

    if values:
        values = _safe_read_yaml(values)
        rv = ValueResolver.get(workflow, values)
        workflow = rv.apply()
        if rv.missing_parameters:
            communication.warn(
                f'Could not resolve the following parameters in "{workflow.name}" workflow: '
                f'{",".join(rv.missing_parameters)}'
            )

    from renku.core.plugins.workflow import workflow_converter

    converter = workflow_converter(format)
    return converter(workflow=workflow, basedir=client.path, output=output, output_format=format)


def export_workflow_command():
    """Command that exports a workflow into a given format."""
    return Command().command(_export_workflow).require_clean().with_database(write=False)


@inject.autoparams()
def _lookup_paths_in_paths(client_dispatcher: IClientDispatcher, lookup_paths: List[str], target_paths: List[str]):
    """Return all lookup_paths that are in or under target_paths."""
    client = client_dispatcher.current_client

    dirs = []
    files = set()

    for p in lookup_paths:
        path = Path(get_relative_paths(client.path, [p])[0])
        if path.is_dir():
            dirs.append(path)
        else:
            files.add(path)

    target_dirs = []
    target_files = set()

    for p in target_paths:
        path = Path(p)
        if path.is_dir():
            target_dirs.append(path)
        else:
            target_files.add(path)

    result = set()

    for target_file in target_files:
        if target_file in files or any(d in target_file.parents for d in dirs):
            result.add(str(target_file))

    for target_dir in target_dirs:
        if target_dir in dirs or any(target_dir in f.parents for f in files):
            result.add(str(target_dir))

    return result


@inject.autoparams()
def _workflow_inputs(activity_gateway: IActivityGateway, paths: List[str] = None):
    """Get inputs used by workflows."""
    usage_paths = activity_gateway.get_all_usage_paths()

    if not paths:
        return usage_paths

    return _lookup_paths_in_paths(lookup_paths=paths, target_paths=usage_paths)


def workflow_inputs_command():
    """Command that shows inputs used by workflows."""
    return Command().command(_workflow_inputs).require_migration().with_database(write=False)


@inject.autoparams()
def _workflow_outputs(activity_gateway: IActivityGateway, paths: List[str] = None):
    """Get inputs used by workflows."""
    generation_paths = activity_gateway.get_all_generation_paths()

    if not paths:
        return generation_paths

    return _lookup_paths_in_paths(lookup_paths=paths, target_paths=generation_paths)


def workflow_outputs_command():
    """Command that shows inputs used by workflows."""
    return Command().command(_workflow_outputs).require_migration().with_database(write=False)


@inject.autoparams()
def execute_workflow(
    plans: List[Plan],
    command_name,
    client_dispatcher: IClientDispatcher,
    activity_gateway: IActivityGateway,
    plan_gateway: IPlanGateway,
    provider="cwltool",
    config=None,
):
    """Execute a Run with/without subprocesses."""
    client = client_dispatcher.current_client

    # NOTE: Pull inputs from Git LFS or other storage backends
    if client.check_external_storage():
        inputs = [i.actual_value for p in plans for i in p.inputs]
        client.pull_paths_from_storage(*inputs)

    delete_indirect_files_list(client.path)

    started_at_time = local_now()

    # NOTE: Create a ``CompositePlan`` because ``workflow_covert`` expects it
    workflow = CompositePlan(id=CompositePlan.generate_id(), plans=plans, name=f"plan-collection-{uuid.uuid4().hex}")
    modified_outputs = execute(workflow=workflow, basedir=client.path, provider=provider, config=config)

    ended_at_time = local_now()

    add_to_git(client.repo.git, *modified_outputs)

    if client.repo.is_dirty():
        postfix = "s" if len(modified_outputs) > 1 else ""
        commit_msg = f"renku {command_name}: committing {len(modified_outputs)} modified file{postfix}"
        committer = Actor(f"renku {__version__}", version_url)
        client.repo.index.commit(commit_msg, committer=committer, skip_hooks=True)

    activities = []

    for plan in plans:
        # NOTE: Update plans are copies of Plan objects. We need to use the original Plan objects to avoid duplicates.
        original_plan = plan_gateway.get_by_id(plan.id)
        activity = Activity.from_plan(plan=original_plan, started_at_time=started_at_time, ended_at_time=ended_at_time)
        activity_gateway.add(activity)
        activities.append(activity)

    if len(activities) > 1:
        activity_collection = ActivityCollection(activities=activities)
        activity_gateway.add_activity_collection(activity_collection)


@inject.autoparams()
def _execute_workflow(
    name_or_id: str, set_params: List[str], provider: str, config: Optional[str], values: Optional[str]
):
    workflow = _find_workflow(name_or_id)

    # apply the provided parameter settings provided by user
    override_params = dict()
    if values:
        override_params.update(_safe_read_yaml(values))

    if set_params:
        for param in set_params:
            name, value = param.split("=", maxsplit=1)
            override_params[name] = value

    if override_params:
        rv = ValueResolver.get(workflow, override_params)
        workflow = rv.apply()

        if rv.missing_parameters:
            communication.warn(
                f'Could not resolve the following parameters in "{workflow.name}" workflow: '
                f'{",".join(rv.missing_parameters)}'
            )

    if config:
        config = _safe_read_yaml(config)

    if isinstance(workflow, CompositePlan):
        import networkx as nx

        graph = ExecutionGraph(workflow=workflow, virtual_links=True)
        plans = list(nx.topological_sort(graph.workflow_graph))
    else:
        plans = [workflow]

    execute_workflow(plans=plans, command_name="execute", provider=provider, config=config)


def execute_workflow_command():
    """Command that executes a workflow."""
    return (
        Command().command(_execute_workflow).require_migration().require_clean().with_database(write=True).with_commit()
    )



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``update`` command."""

import os
from pathlib import Path

import git

from renku.cli import cli
from renku.core.management.repository import DEFAULT_DATA_DIR as DATA_DIR
from renku.core.metadata.gateway.activity_gateway import ActivityGateway
from renku.core.models.workflow.plan import Plan
from tests.utils import format_result_exception, write_and_commit_file


def test_update(runner, client, renku_cli, client_database_injection_manager):
    """Test output is updated when source changes."""
    repo = client.repo
    source = os.path.join(client.path, "source.txt")
    output = os.path.join(client.path, "output.txt")

    write_and_commit_file(repo, source, "content")

    exit_code, previous_activity = renku_cli("run", "head", "-1", source, stdout=output)
    assert 0 == exit_code

    write_and_commit_file(repo, source, "changed content")

    exit_code, activity = renku_cli("update", "--all")

    assert 0 == exit_code
    plan = activity.association.plan
    assert previous_activity.association.plan.id == plan.id
    assert isinstance(plan, Plan)

    assert "changed content" == Path(output).read_text()

    result = runner.invoke(cli, ["status"])
    assert 0 == result.exit_code, format_result_exception(result)

    with client_database_injection_manager(client):
        activity_gateway = ActivityGateway()
        activity_collections = activity_gateway.get_all_activity_collections()

        # NOTE: No ActivityCollection is created if update include only one activity
        assert [] == activity_collections


def test_update_multiple_steps(runner, client, renku_cli, client_database_injection_manager):
    """Test update in a multi-step workflow."""
    repo = client.repo
    source = os.path.join(client.path, "source.txt")
    intermediate = os.path.join(client.path, "intermediate.txt")
    output = os.path.join(client.path, "output.txt")

    write_and_commit_file(repo, source, "content")

    exit_code, activity1 = renku_cli("run", "cp", source, intermediate)
    assert 0 == exit_code
    exit_code, activity2 = renku_cli("run", "cp", intermediate, output)
    assert 0 == exit_code

    write_and_commit_file(repo, source, "changed content")

    exit_code, activities = renku_cli("update", "--all")

    assert 0 == exit_code
    plans = [a.association.plan for a in activities]
    assert 2 == len(plans)
    assert isinstance(plans[0], Plan)
    assert isinstance(plans[1], Plan)
    assert {p.id for p in plans} == {activity1.association.plan.id, activity2.association.plan.id}

    assert "changed content" == Path(intermediate).read_text()
    assert "changed content" == Path(output).read_text()

    result = runner.invoke(cli, ["status"])
    assert 0 == result.exit_code, format_result_exception(result)

    with client_database_injection_manager(client):
        activity_gateway = ActivityGateway()
        activity_collections = activity_gateway.get_all_activity_collections()

        assert 1 == len(activity_collections)
        assert {a.id for a in activities} == {a.id for a in activity_collections[0].activities}


def test_update_multiple_steps_with_path(runner, project, renku_cli):
    """Test update in a multi-step workflow when a path is specified."""
    repo = git.Repo(project)
    source = os.path.join(project, "source.txt")
    intermediate = os.path.join(project, "intermediate.txt")
    output = os.path.join(project, "output.txt")

    write_and_commit_file(repo, source, "content")

    exit_code, activity1 = renku_cli("run", "cp", source, intermediate)
    assert 0 == exit_code
    exit_code, _ = renku_cli("run", "cp", intermediate, output)
    assert 0 == exit_code

    write_and_commit_file(repo, source, "changed content")

    exit_code, activity = renku_cli("update", intermediate)

    assert 0 == exit_code
    plan = activity.association.plan
    assert isinstance(plan, Plan)
    assert plan.id == activity1.association.plan.id

    assert "changed content" == Path(intermediate).read_text()
    assert "content" == Path(output).read_text()

    result = runner.invoke(cli, ["status"])
    assert 1 == result.exit_code, format_result_exception(result)
    assert "output.txt: intermediate.txt" in result.output
    assert "source.txt" not in result.output


def test_update_with_directory_paths(project, renku_cli):
    """Test update when a directory path is specified."""
    repo = git.Repo(project)
    data = os.path.join(project, "data", "dataset", "my-data")
    Path(data).mkdir(parents=True, exist_ok=True)
    source = os.path.join(project, "source.txt")
    output = os.path.join(data, "output.txt")

    write_and_commit_file(repo, source, "content")

    exit_code, previous_activity = renku_cli("run", "head", "-1", source, stdout=output)
    assert 0 == exit_code

    write_and_commit_file(repo, source, "changed content")

    exit_code, activity = renku_cli("update", data)

    assert 0 == exit_code
    assert "changed content" == Path(output).read_text()
    plan = activity.association.plan
    assert previous_activity.association.plan.id == plan.id


def test_multiple_updates(runner, project, renku_cli):
    """Test multiple updates of the same source."""
    repo = git.Repo(project)
    source = os.path.join(project, "source.txt")
    intermediate = os.path.join(project, "intermediate.txt")
    output = os.path.join(project, "output.txt")

    write_and_commit_file(repo, source, "content")

    exit_code, activity1 = renku_cli("run", "cp", source, intermediate)
    assert 0 == exit_code
    exit_code, activity2 = renku_cli("run", "cp", intermediate, output)
    assert 0 == exit_code

    write_and_commit_file(repo, source, "changed content")

    exit_code, _ = renku_cli("update", "--all")
    assert 0 == exit_code
    assert "changed content" == Path(intermediate).read_text()

    write_and_commit_file(repo, source, "more changed content")

    exit_code, activities = renku_cli("update", "--all")

    assert 0 == exit_code
    plans = [a.association.plan for a in activities]
    assert 2 == len(plans)
    assert isinstance(plans[0], Plan)
    assert isinstance(plans[1], Plan)
    assert {p.id for p in plans} == {activity1.association.plan.id, activity2.association.plan.id}

    assert "more changed content" == Path(intermediate).read_text()
    assert "more changed content" == Path(output).read_text()

    result = runner.invoke(cli, ["status"])
    assert 0 == result.exit_code, format_result_exception(result)


def test_update_workflow_without_outputs(runner, project, run):
    """Test workflow without outputs."""
    repo = git.Repo(project)
    source = os.path.join(project, "source.txt")

    write_and_commit_file(repo, source, "content")

    assert 0 == runner.invoke(cli, ["run", "cat", "--no-output", source]).exit_code

    write_and_commit_file(repo, source, "changes")

    assert 1 == runner.invoke(cli, ["status"]).exit_code

    assert 0 == run(args=["update", "--all"])

    result = runner.invoke(cli, ["status"])

    # NOTE: Activity is updated or otherwise status would still return 1
    assert 0 == result.exit_code, format_result_exception(result)


def test_update_siblings(project, run, no_lfs_warning):
    """Test all generations of an activity are updated together."""
    repo = git.Repo(project)
    parent = os.path.join(project, "parent.txt")
    brother = os.path.join(project, "brother.txt")
    sister = os.path.join(project, "sister.txt")
    siblings = [Path(brother), Path(sister)]

    write_and_commit_file(repo, parent, "content")

    assert 0 == run(args=["run", "tee", brother, sister], stdin=parent)

    # The output file is copied from the source.
    for sibling in siblings:
        assert "content" == sibling.read_text()

    write_and_commit_file(repo, parent, "changed content")

    assert 0 == run(args=["update", brother])

    for sibling in siblings:
        assert "changed content" == sibling.read_text()

    # Siblings kept together even when one is removed.
    repo.index.remove([brother], working_tree=True)
    repo.index.commit("Brother removed")
    assert not os.path.exists(brother)

    write_and_commit_file(repo, parent, "more content")

    # Update should create the missing sibling
    assert 0 == run(args=["update", "--all"])

    for sibling in siblings:
        assert "more content" == sibling.read_text()


def test_update_siblings_in_output_directory(project, run):
    """Files in output directory are linked or removed after update."""
    repo = git.Repo(project)
    source = os.path.join(project, "source.txt")
    output = Path(os.path.join(project, "output"))  # a directory

    def write_source():
        """Write source from files."""
        write_and_commit_file(repo, source, content="\n".join(" ".join(line) for line in files) + "\n")

    def check_files():
        """Check file content."""
        assert len(files) == len(list(output.rglob("*")))

        for name, content in files:
            assert content == (output / name).read_text().strip()

    files = [("first", "1"), ("second", "2"), ("third", "3")]
    write_source()

    script = 'mkdir -p "$0"; ' "cat - | while read -r name content; do " 'echo "$content" > "$0/$name"; done'

    assert not os.path.exists(output)

    assert 0 == run(args=["run", "sh", "-c", script, "output"], stdin=source)

    assert os.path.exists(output)
    check_files()

    files = [("third", "3"), ("fourth", "4")]
    write_source()

    assert 0 == run(args=["update", "output"])

    check_files()


def test_update_relative_path_for_directory_input(client, run, renku_cli):
    """Test having a directory input generates relative path in CWL."""
    write_and_commit_file(client.repo, client.path / DATA_DIR / "file1", "file1")

    assert 0 == run(args=["run", "ls", DATA_DIR], stdout="ls.data")

    write_and_commit_file(client.repo, client.path / DATA_DIR / "file2", "file2")

    exit_code, activity = renku_cli("update", "--all")

    assert 0 == exit_code
    plan = activity.association.plan
    assert 1 == len(plan.inputs)
    assert "data" == plan.inputs[0].default_value


def test_update_no_args(runner, project, no_lfs_warning):
    """Test calling update with no args raises ParameterError."""
    repo = git.Repo(project)
    source = os.path.join(project, "source.txt")
    output = os.path.join(project, "output.txt")

    write_and_commit_file(repo, source, "content")

    assert 0 == runner.invoke(cli, ["run", "cp", source, output]).exit_code

    write_and_commit_file(repo, source, "changed content")

    before_commit = repo.head.commit

    result = runner.invoke(cli, ["update"])

    assert 2 == result.exit_code
    assert "Either PATHS, --all/-a, or --dry-run/-n should be specified." in result.output

    assert before_commit == repo.head.commit


def test_update_with_no_execution(project, runner):
    """Test update when no workflow is executed."""
    repo = git.Repo(project)
    input = os.path.join(project, "data", "input.txt")
    write_and_commit_file(repo, input, "content")

    result = runner.invoke(cli, ["update", input], catch_exceptions=False)

    assert 1 == result.exit_code


def test_update_overridden_output(project, renku_cli, runner):
    """Test a path where final output is overridden will be updated partially."""
    repo = git.Repo(project)
    a = os.path.join(project, "a")
    b = os.path.join(project, "b")
    c = os.path.join(project, "c")

    write_and_commit_file(repo, a, "content")

    assert 0 == runner.invoke(cli, ["run", "--name", "r1", "cp", a, b]).exit_code
    assert 0 == runner.invoke(cli, ["run", "--name", "r2", "cp", b, c]).exit_code
    assert 0 == renku_cli("run", "--name", "r3", "wc", a, stdout=c).exit_code

    write_and_commit_file(repo, a, "new content")

    result = runner.invoke(cli, ["update", "--dry-run"])

    assert 0 == result.exit_code
    assert "r1" in result.output
    assert "r2" not in result.output
    assert "r3" in result.output


def test_update_overridden_outputs_partially(project, renku_cli, runner):
    """Test a path where one of the final output is overridden will be updated completely but in proper order."""
    repo = git.Repo(project)
    a = os.path.join(project, "a")
    b = os.path.join(project, "b")
    c = os.path.join(project, "c")
    d = os.path.join(project, "d")

    write_and_commit_file(repo, a, "content")

    assert 0 == runner.invoke(cli, ["run", "--name", "r1", "cp", a, b]).exit_code
    assert 0 == renku_cli("run", "--name", "r2", "tee", c, d, stdin=b).exit_code
    assert 0 == renku_cli("run", "--name", "r3", "wc", a, stdout=c).exit_code

    write_and_commit_file(repo, a, "new content")

    result = runner.invoke(cli, ["update", "--dry-run"])

    assert 0 == result.exit_code
    assert "r1" in result.output
    assert "r2" in result.output
    assert "r3" in result.output
    assert result.output.find("r2") < result.output.find("r3")


def test_update_multiple_paths_common_output(project, renku_cli, runner):
    """Test multiple paths that generate the same output will be updated except the last overridden step."""
    repo = git.Repo(project)
    a = os.path.join(project, "a")
    b = os.path.join(project, "b")
    c = os.path.join(project, "c")
    d = os.path.join(project, "d")

    write_and_commit_file(repo, a, "content")

    assert 0 == runner.invoke(cli, ["run", "--name", "r1", "cp", a, b]).exit_code
    assert 0 == runner.invoke(cli, ["run", "--name", "r2", "cp", b, d]).exit_code
    assert 0 == runner.invoke(cli, ["run", "--name", "r3", "cp", a, c]).exit_code
    assert 0 == renku_cli("run", "--name", "r4", "wc", c, stdout=d).exit_code

    write_and_commit_file(repo, a, "new content")

    result = runner.invoke(cli, ["update", "--dry-run"])

    assert 0 == result.exit_code
    assert "r1" in result.output
    assert "r2" not in result.output
    assert "r3" in result.output
    assert "r4" in result.output



# -*- coding: utf-8 -*-
#
# Copyright 2020-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service templates view tests."""
import base64
import json
from copy import deepcopy
from io import BytesIO
from pathlib import Path
from tempfile import TemporaryDirectory
from time import sleep

import pytest

from renku.core.commands.init import fetch_template_from_git, read_template_manifest
from renku.core.utils.scm import normalize_to_ascii
from renku.service.config import RENKU_EXCEPTION_ERROR_CODE
from tests.utils import retry_failed


@pytest.mark.service
@pytest.mark.integration
@retry_failed
def test_read_manifest_from_template(svc_client_with_templates):
    """Check reading manifest template."""
    from PIL import Image

    svc_client, headers, template_params = svc_client_with_templates

    response = svc_client.get("/templates.read_manifest", query_string=template_params, headers=headers)

    assert response
    assert {"result"} == set(response.json.keys())
    assert response.json["result"]["templates"]

    templates = response.json["result"]["templates"]
    assert len(templates) > 0

    default_template = templates[template_params["index"] - 1]
    assert default_template["folder"] == template_params["id"]
    assert "icon" in default_template and default_template["icon"]
    icon = Image.open(BytesIO(base64.b64decode(default_template["icon"])))
    assert icon.size == (256, 256)


@pytest.mark.service
@pytest.mark.integration
@retry_failed
def test_compare_manifests(svc_client_with_templates):
    """Check reading manifest template."""
    svc_client, headers, template_params = svc_client_with_templates

    response = svc_client.get("/templates.read_manifest", query_string=template_params, headers=headers)

    assert response
    assert {"result"} == set(response.json.keys())
    assert response.json["result"]["templates"]

    with TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)
        manifest_file, _ = fetch_template_from_git(template_params["url"], template_params["ref"], temp_path)
        manifest = read_template_manifest(temp_path)

        assert manifest_file and manifest_file.exists()
        assert manifest

        templates_service = response.json["result"]["templates"]
        templates_local = manifest
        default_index = template_params["index"] - 1

        if "icon" in templates_service[default_index]:
            del templates_service[default_index]["icon"]
        if "icon" in templates_local[default_index]:
            del templates_local[default_index]["icon"]

        assert templates_service[default_index] == templates_local[default_index]


@pytest.mark.service
@pytest.mark.integration
@retry_failed
def test_create_project_from_template(svc_client_templates_creation):
    """Check reading manifest template."""
    from git import Repo

    from renku.service.serializers.headers import RenkuHeaders
    from renku.service.utils import CACHE_PROJECTS_PATH

    svc_client, headers, payload, rm_remote = svc_client_templates_creation

    # NOTE:  fail: remote authentication
    anonymous_headers = deepcopy(headers)
    anonymous_headers["Authorization"] = "Bearer None"
    response = svc_client.post("/templates.create_project", data=json.dumps(payload), headers=anonymous_headers)

    assert response
    assert response.json["error"]
    assert "Cannot push changes" in response.json["error"]["reason"]

    # NOTE:  fail: missing parameters
    if len(payload["parameters"]) > 0:
        payload_without_parameters = deepcopy(payload)
        payload_without_parameters["parameters"] = []
        response = svc_client.post(
            "/templates.create_project", data=json.dumps(payload_without_parameters), headers=headers
        )
        assert response
        assert response.json["error"]
        assert RENKU_EXCEPTION_ERROR_CODE == response.json["error"]["code"]
        assert "missing parameter" in response.json["error"]["reason"]

    # NOTE:  successfully push with proper authentication
    response = svc_client.post("/templates.create_project", data=json.dumps(payload), headers=headers)

    assert response
    assert {"result"} == set(response.json.keys())
    stripped_name = normalize_to_ascii(payload["project_name"])
    assert stripped_name == response.json["result"]["slug"]
    expected_url = "{0}/{1}/{2}".format(payload["project_repository"], payload["project_namespace"], stripped_name)
    assert expected_url == response.json["result"]["url"]

    # NOTE: assert correct git user is set on new project
    user_data = RenkuHeaders.decode_user(headers["Renku-User"])
    project_path = (
        CACHE_PROJECTS_PATH
        / user_data["user_id"]
        / response.json["result"]["project_id"]
        / payload["project_namespace"]
        / stripped_name
    )
    repo = Repo(project_path)
    reader = repo.config_reader()
    assert reader.get_value("user", "email") == user_data["email"]
    assert reader.get_value("user", "name") == user_data["name"]

    # NOTE:  successfully re-use old name after cleanup
    assert rm_remote() is True
    sleep(1)  # NOTE: sleep to make sure remote isn't locked
    response = svc_client.post("/templates.create_project", data=json.dumps(payload), headers=headers)
    assert response
    assert {"result"} == set(response.json.keys())
    assert expected_url == response.json["result"]["url"]
    assert rm_remote() is True



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test activity database gateways."""

from datetime import datetime, timedelta

from renku.core.metadata.gateway.activity_gateway import ActivityGateway
from renku.core.models.provenance.activity import Activity, Association
from renku.core.models.workflow.plan import Plan
from tests.utils import create_dummy_activity


def test_activity_gateway_get_latest_activity(dummy_database_injection_manager):
    """Test getting latest activity for a plan."""

    plan = Plan(id=Plan.generate_id(), name="plan")
    plan2 = Plan(id=Plan.generate_id(), name="plan2")

    activity1_id = Activity.generate_id()
    activity1 = Activity(
        id=activity1_id,
        ended_at_time=datetime.utcnow() - timedelta(hours=1),
        association=Association(id=Association.generate_id(activity1_id), plan=plan),
    )

    activity2_id = Activity.generate_id()
    activity2 = Activity(
        id=activity2_id,
        ended_at_time=datetime.utcnow(),
        association=Association(id=Association.generate_id(activity2_id), plan=plan),
    )

    activity3_id = Activity.generate_id()
    activity3 = Activity(
        id=activity3_id,
        ended_at_time=datetime.utcnow(),
        association=Association(id=Association.generate_id(activity3_id), plan=plan2),
    )

    with dummy_database_injection_manager(None):
        activity_gateway = ActivityGateway()

        activity_gateway.add(activity1)

        latest_activities = activity_gateway.get_latest_activity_per_plan()
        assert len(latest_activities) == 1
        assert {activity1_id} == {a.id for a in latest_activities.values()}
        assert {plan.id} == {p.id for p in latest_activities.keys()}

        activity_gateway.add(activity3)

        latest_activities = activity_gateway.get_latest_activity_per_plan()
        assert len(latest_activities) == 2
        assert {activity1_id, activity3_id} == {a.id for a in latest_activities.values()}
        assert {plan.id, plan2.id} == {p.id for p in latest_activities.keys()}

        activity_gateway.add(activity2)

        latest_activities = activity_gateway.get_latest_activity_per_plan()
        assert len(latest_activities) == 2
        assert {activity2_id, activity3_id} == {a.id for a in latest_activities.values()}
        assert {plan.id, plan2.id} == {p.id for p in latest_activities.keys()}


def test_activity_gateway_get_latest_plan_usages(dummy_database_injection_manager):
    """Test getting latest activity for a plan."""

    plan = Plan(id=Plan.generate_id(), name="plan")
    plan2 = Plan(id=Plan.generate_id(), name="plan2")

    activity1 = create_dummy_activity(plan=plan, usages=["in1"])
    activity2 = create_dummy_activity(plan=plan, usages=["in1"])
    activity3 = create_dummy_activity(plan=plan2, usages=["in2"])

    with dummy_database_injection_manager(None):
        activity_gateway = ActivityGateway()

        activity_gateway.add(activity1)
        activity_gateway.add(activity2)
        activity_gateway.add(activity3)

        latest_usages = activity_gateway.get_plans_and_usages_for_latest_activities()

        assert len(latest_usages) == 2
        assert latest_usages[plan] == activity2.usages
        assert latest_usages[plan2] == activity3.usages


def test_activity_gateway_downstream_activities(dummy_database_injection_manager):
    """test getting downstream activities work."""
    plan = Plan(id=Plan.generate_id(), name="plan")

    intermediate = create_dummy_activity(plan=plan, usages=["some/data"], generations=["other/data/file"])
    previous = create_dummy_activity(plan=plan, generations=["some/"])
    following = create_dummy_activity(plan=plan, usages=["other/data"])
    unrelated = create_dummy_activity(plan=plan, usages=["unrelated_in"], generations=["unrelated_out"])

    with dummy_database_injection_manager(None):
        activity_gateway = ActivityGateway()

        activity_gateway.add(intermediate)
        activity_gateway.add(following)
        activity_gateway.add(previous)
        activity_gateway.add(unrelated)

        downstream = activity_gateway.get_downstream_activities(following)

        assert not downstream

        downstream = activity_gateway.get_downstream_activities(intermediate)
        assert {following.id} == {a.id for a in downstream}

        downstream = activity_gateway.get_downstream_activities(previous)
        assert {following.id, intermediate.id} == {a.id for a in downstream}


def test_activity_gateway_downstream_activity_chains(dummy_database_injection_manager):
    """test getting downstream activity chains work."""
    r1 = create_dummy_activity(plan="r1", usages=["a"], generations=["b"])
    r2 = create_dummy_activity(plan="r2", usages=["b"], generations=["c"])
    r3 = create_dummy_activity(plan="r3", usages=["d"], generations=["e"])
    r4 = create_dummy_activity(plan="r4", usages=["c", "e"], generations=["f", "g"])
    r5 = create_dummy_activity(plan="r5", usages=["f"], generations=["h"])
    r6 = create_dummy_activity(plan="r6", usages=["g"], generations=["i"])
    r7 = create_dummy_activity(plan="r7", usages=["x"], generations=["y"])

    with dummy_database_injection_manager(None):
        activity_gateway = ActivityGateway()

        activity_gateway.add(r1)
        activity_gateway.add(r3)
        activity_gateway.add(r2)
        activity_gateway.add(r4)
        activity_gateway.add(r5)
        activity_gateway.add(r6)
        activity_gateway.add(r7)

        assert [] == activity_gateway.get_downstream_activity_chains(r6)

        downstream_chains = activity_gateway.get_downstream_activity_chains(r1)
        assert {(r2.id,), (r2.id, r4.id), (r2.id, r4.id, r5.id), (r2.id, r4.id, r6.id)} == {
            tuple(a.id for a in chain) for chain in downstream_chains
        }

        downstream_chains = activity_gateway.get_downstream_activity_chains(r4)
        assert {(r5.id,), (r6.id,)} == {tuple(a.id for a in chain) for chain in downstream_chains}

        assert [] == activity_gateway.get_downstream_activity_chains(r7)


def test_activity_gateway_upstream_activity_chains(dummy_database_injection_manager):
    """test getting upstream activity chains work."""
    r1 = create_dummy_activity(plan="r1", usages=["a"], generations=["b"])
    r2 = create_dummy_activity(plan="r2", usages=["b"], generations=["c"])
    r3 = create_dummy_activity(plan="r3", usages=["d"], generations=["e"])
    r4 = create_dummy_activity(plan="r4", usages=["c", "e"], generations=["f", "g"])
    r5 = create_dummy_activity(plan="r5", usages=["f"], generations=["h"])
    r6 = create_dummy_activity(plan="r6", usages=["g"], generations=["i"])
    r7 = create_dummy_activity(plan="r7", usages=["x"], generations=["y"])

    with dummy_database_injection_manager(None):
        activity_gateway = ActivityGateway()

        activity_gateway.add(r1)
        activity_gateway.add(r3)
        activity_gateway.add(r2)
        activity_gateway.add(r4)
        activity_gateway.add(r5)
        activity_gateway.add(r6)
        activity_gateway.add(r7)

        assert [] == activity_gateway.get_upstream_activity_chains(r1)

        downstream_chains = activity_gateway.get_upstream_activity_chains(r6)
        assert {(r4.id,), (r4.id, r3.id), (r4.id, r2.id), (r4.id, r2.id, r1.id)} == {
            tuple(a.id for a in chain) for chain in downstream_chains
        }

        downstream_chains = activity_gateway.get_upstream_activity_chains(r4)
        assert {(r3.id,), (r2.id,), (r2.id, r1.id)} == {tuple(a.id for a in chain) for chain in downstream_chains}

        assert [] == activity_gateway.get_upstream_activity_chains(r7)



# -*- coding: utf-8 -*-
#
# Copyright 2021 Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku fixtures for metadata Database."""

import copy
import datetime
from typing import Tuple

import pytest

from renku.core import errors
from renku.core.management.interface.database_dispatcher import IDatabaseDispatcher
from renku.core.metadata.database import Database
from renku.core.metadata.gateway.database_gateway import initialize_database


class DummyStorage:
    """An in-memory storage class."""

    def __init__(self):
        self._files = {}
        self._modification_dates = {}

    def store(self, filename: str, data, compress=False):
        """Store object."""
        assert isinstance(filename, str)

        self._files[filename] = data
        self._modification_dates[filename] = datetime.datetime.now()

    def load(self, filename: str):
        """Load data for object with object id oid."""
        assert isinstance(filename, str)

        if filename not in self._files:
            raise errors.ObjectNotFoundError(filename)

        return copy.deepcopy(self._files[filename])

    def get_modification_date(self, filename: str):
        """Return modification date of a file."""
        return self._modification_dates[filename]

    def exists(self, filename: str):
        """Return True if filename exists in the storage."""
        return filename in self._files


class DummyDatabaseDispatcher:
    """DatabaseDispatcher with DummyStorage.

    Handles getting current database (Database) and entering/exiting the stack for the database.
    """

    def __init__(self, database: Database):
        self.database = database

    @property
    def current_database(self) -> Database:
        """Get the currently active database."""
        return self.database

    def push_database_to_stack(self, path: str, commit: bool = False) -> None:
        """Create and push a new database to the stack."""
        pass

    def pop_database(self) -> None:
        """Remove the current database from the stack."""
        pass

    def finalize_dispatcher(self) -> None:
        """Close all database contexts."""
        pass


@pytest.fixture
def database() -> Tuple[Database, DummyStorage]:
    """A Database with in-memory storage."""
    storage = DummyStorage()
    database = Database(storage=storage)

    initialize_database(database)

    yield database, storage


@pytest.fixture
def database_injection_bindings():
    """Create injection bindings for a database."""

    def _add_database_injection_bindings(bindings):
        from renku.core.management.command_builder.database_dispatcher import DatabaseDispatcher
        from renku.core.management.interface.activity_gateway import IActivityGateway
        from renku.core.management.interface.client_dispatcher import IClientDispatcher
        from renku.core.management.interface.database_dispatcher import IDatabaseDispatcher
        from renku.core.management.interface.database_gateway import IDatabaseGateway
        from renku.core.management.interface.dataset_gateway import IDatasetGateway
        from renku.core.management.interface.plan_gateway import IPlanGateway
        from renku.core.management.interface.project_gateway import IProjectGateway
        from renku.core.metadata.gateway.activity_gateway import ActivityGateway
        from renku.core.metadata.gateway.database_gateway import DatabaseGateway
        from renku.core.metadata.gateway.dataset_gateway import DatasetGateway
        from renku.core.metadata.gateway.plan_gateway import PlanGateway
        from renku.core.metadata.gateway.project_gateway import ProjectGateway

        dispatcher = DatabaseDispatcher()
        dispatcher.push_database_to_stack(
            bindings["bindings"][IClientDispatcher].current_client.database_path, commit=True
        )

        bindings["bindings"][IDatabaseDispatcher] = dispatcher

        bindings["constructor_bindings"][IPlanGateway] = lambda: PlanGateway()
        bindings["constructor_bindings"][IActivityGateway] = lambda: ActivityGateway()
        bindings["constructor_bindings"][IDatabaseGateway] = lambda: DatabaseGateway()
        bindings["constructor_bindings"][IDatasetGateway] = lambda: DatasetGateway()
        bindings["constructor_bindings"][IProjectGateway] = lambda: ProjectGateway()

        return bindings

    return _add_database_injection_bindings


@pytest.fixture
def dummy_database_injection_bindings(database):
    """Create injection bindings for a database."""

    def _add_database_injection_bindings(bindings):
        from renku.core.management.interface.activity_gateway import IActivityGateway
        from renku.core.management.interface.database_gateway import IDatabaseGateway
        from renku.core.management.interface.dataset_gateway import IDatasetGateway
        from renku.core.management.interface.plan_gateway import IPlanGateway
        from renku.core.management.interface.project_gateway import IProjectGateway
        from renku.core.metadata.gateway.activity_gateway import ActivityGateway
        from renku.core.metadata.gateway.database_gateway import DatabaseGateway
        from renku.core.metadata.gateway.dataset_gateway import DatasetGateway
        from renku.core.metadata.gateway.plan_gateway import PlanGateway
        from renku.core.metadata.gateway.project_gateway import ProjectGateway

        bindings["bindings"][IDatabaseDispatcher] = DummyDatabaseDispatcher(database[0])

        bindings["constructor_bindings"][IPlanGateway] = lambda: PlanGateway()
        bindings["constructor_bindings"][IActivityGateway] = lambda: ActivityGateway()
        bindings["constructor_bindings"][IDatabaseGateway] = lambda: DatabaseGateway()
        bindings["constructor_bindings"][IDatasetGateway] = lambda: DatasetGateway()
        bindings["constructor_bindings"][IProjectGateway] = lambda: ProjectGateway()

        return bindings

    return _add_database_injection_bindings


@pytest.fixture
def injected_client_with_database(client, client_injection_bindings, database_injection_bindings, injection_binder):
    """Inject a client."""
    bindings = database_injection_bindings(client_injection_bindings(client))
    injection_binder(bindings)


@pytest.fixture
def injected_local_client_with_database(
    local_client, client_injection_bindings, database_injection_bindings, injection_binder
):
    """Inject a client."""
    bindings = database_injection_bindings(client_injection_bindings(local_client))
    injection_binder(bindings)


@pytest.fixture
def injection_manager():
    """Factory fixture for injection manager."""

    def _injection_manager(bindings):
        from tests.utils import injection_manager

        return injection_manager(bindings)

    return _injection_manager


@pytest.fixture
def client_database_injection_manager(client_injection_bindings, database_injection_bindings, injection_manager):
    """Fixture for context manager with client and db injection."""

    def _inner(client):
        return injection_manager(database_injection_bindings(client_injection_bindings(client)))

    return _inner


@pytest.fixture
def dummy_database_injection_manager(dummy_database_injection_bindings, injection_manager):
    """Fixture for context manager with client and db injection."""

    def _inner(client):
        return injection_manager(dummy_database_injection_bindings({"bindings": {}, "constructor_bindings": {}}))

    return _inner


@pytest.fixture
def injected_client_with_dummy_database(
    client, client_injection_bindings, dummy_database_injection_bindings, injection_binder
):
    """Inject a client."""
    bindings = dummy_database_injection_bindings(client_injection_bindings(client))
    injection_binder(bindings)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Plugin hooks for renku run customization."""
from pathlib import Path
from typing import Any, Dict, List, Tuple

import pluggy

from renku.core import errors
from renku.core.models.workflow.plan import AbstractPlan
from renku.core.models.workflow.provider import IWorkflowProvider

hookspec = pluggy.HookspecMarker("renku")


@hookspec
def workflow_provider() -> Tuple[IWorkflowProvider, str]:
    """Plugin Hook to get providers for ``workflow execute`` call.

    :returns: A tuple of the provider itself and the workflow executor backends name.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def workflow_execute(workflow: AbstractPlan, basedir: Path, config: Dict[str, Any]):
    """Plugin Hook for ``workflow execute`` call.

    Can be used to execute renku workflows with different workflow executors.

    :param workflow: a ``AbstractPlan`` object that describes the given workflow.
    :param config: a configuration for the provider.
    """
    raise NotImplementedError


def available_workflow_providers() -> List[str]:
    """Returns the currently available workflow execution providers."""
    from renku.core.plugins.pluginmanager import get_plugin_manager

    pm = get_plugin_manager()
    providers = pm.hook.workflow_provider()
    return [p[1] for p in providers]


def execute(workflow: AbstractPlan, basedir: Path, config: Dict[str, Any], provider: str = "cwltool") -> List[str]:
    """Executes a given workflow using the selected provider.

    :param workflow: Workflow to be executed.
    :param basedir: The root directory of the renku project.
    :param config: Configuration values for the workflow provider.
    :param provider: The workflow executor engine to be used.
    :returns: List of paths that has been modified.
    """
    from renku.core.plugins.pluginmanager import get_plugin_manager

    pm = get_plugin_manager()
    providers = pm.hook.workflow_provider()
    provider = next(filter(lambda x: provider == x[1], providers), None)
    if not provider:
        raise errors.ParameterError(f"The specified workflow executor '{provider}' is not available.")

    providers.remove(provider)
    executor = pm.subset_hook_caller("workflow_execute", list(map(lambda x: x[0], providers)))

    return executor(workflow=workflow, basedir=basedir, config=config)



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Represent provenance entities."""

import os.path
from pathlib import Path
from typing import List, Union
from urllib.parse import quote

from renku.core.metadata.immutable import Immutable
from renku.core.models.calamus import JsonLDSchema, Nested, fields, prov, renku
from renku.core.utils.git import get_object_hash

_entity_cache = {}


class Entity(Immutable):
    """Represent a file."""

    __slots__ = ("checksum", "path")

    def __init__(self, *, checksum: str, id: str = None, path: Union[Path, str], **kwargs):
        path = str(path)

        assert id is None or isinstance(id, str)
        assert not os.path.isabs(path), f"Entity is being created with absolute path: '{path}'"

        id = id or Entity.generate_id(checksum, path)
        super().__init__(checksum=checksum, id=id, path=path, **kwargs)

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, Entity):
            return False
        return self.checksum == other.checksum and self.path == other.path

    def __hash__(self):
        return hash((self.checksum, self.path))

    @staticmethod
    def generate_id(checksum: str, path: Union[Path, str]) -> str:
        """Generate an Entity identifier."""
        quoted_path = quote(str(path).strip("/"))

        return f"/entities/{checksum}/{quoted_path}"

    @classmethod
    def from_revision(cls, client, path: Union[Path, str], revision: str = None) -> "Entity":
        """Return dependency from given path and revision."""
        revision = revision or "HEAD"
        assert isinstance(revision, str), f"Invalid revision: {revision}"

        # FIXME: Implement the cache properly: It's mainly used in migrations
        global _entity_cache
        key = (revision, str(path))
        cached_entry = _entity_cache.get(key)
        if cached_entry:
            return cached_entry

        # TODO: What checksum we get at "HEAD" if object is staged but not committed
        checksum = get_object_hash(repo=client.repo, revision=revision, path=path)
        # NOTE: If object was not found at a revision it's either removed or exists in a different revision; keep the
        # entity and use revision as checksum
        checksum = checksum or revision
        # TODO: What would be checksum for a directory if it's not committed yet.
        id = cls.generate_id(checksum=checksum, path=path)

        absolute_path = client.path / path
        if str(path) != "." and absolute_path.is_dir():
            members = cls.get_directory_members(client, revision, absolute_path)
            entity = Collection(id=id, checksum=checksum, path=path, members=members)
        else:
            entity = cls(id=id, checksum=checksum, path=path)

        _entity_cache[cached_entry] = entity

        return entity

    @classmethod
    def get_directory_members(cls, client, revision, absolute_path: Path) -> List["Entity"]:
        """Return first-level files/directories in a directory."""
        members: List[Entity] = []

        for member in absolute_path.iterdir():
            if member.name == ".gitkeep":
                continue

            member_path = str(member.relative_to(client.path))

            try:
                assert all(member_path != m.path for m in members)

                entity = cls.from_revision(client, member_path, revision)
                # NOTE: If a path is not found at a revision we assume that it didn't exist at that revision
                if entity:
                    members.append(entity)
            except KeyError:
                pass

        return members


class Collection(Entity):
    """Represent a directory with files."""

    __slots__ = ("members",)

    def __init__(self, *, checksum: str, id: str = None, path: Union[Path, str], members: List[Entity] = None):
        members = tuple(members) if members else ()
        super().__init__(checksum=checksum, id=id, path=path, members=members)


class EntitySchema(JsonLDSchema):
    """Entity Schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Entity]
        model = Entity

    checksum = fields.String(renku.checksum, missing=None)
    id = fields.Id()
    path = fields.String(prov.atLocation)


class CollectionSchema(EntitySchema):
    """Entity Schema."""

    class Meta:
        """Meta class."""

        rdf_type = prov.Collection
        model = Collection

    members = Nested(prov.hadMember, [EntitySchema, "CollectionSchema"], many=True)



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Custom database for store Persistent objects."""

import datetime
import gzip
import hashlib
import json
from pathlib import Path
from types import BuiltinFunctionType, FunctionType
from typing import Dict, List, Optional, Union
from uuid import uuid4

import persistent
from BTrees.OOBTree import OOBTree
from persistent import GHOST, UPTODATE
from persistent.interfaces import IPickleCache
from ZODB.utils import z64
from zope.interface import implementer
from zope.interface.interface import InterfaceClass

from renku.core import errors
from renku.core.metadata.immutable import Immutable

OID_TYPE = str
TYPE_TYPE = "type"
FUNCTION_TYPE = "function"
REFERENCE_TYPE = "reference"
MARKER = object()

"""NOTE: These are used as _p_serial to mark if an object was read from storage or is new"""
NEW = z64  # NOTE: Do not change this value since this is the default when a Persistent object is created
PERSISTED = b"1" * 8


def get_type_name(object) -> Optional[str]:
    """Return fully-qualified object's type name."""
    if object is None:
        return None

    object_type = object if isinstance(object, type) else type(object)
    return f"{object_type.__module__}.{object_type.__qualname__}"


def get_class(type_name: Optional[str]) -> Optional[type]:
    """Return the class for a fully-qualified type name."""
    if type_name is None:
        return None

    components = type_name.split(".")
    module_name = components[0]

    if module_name not in ["BTrees", "builtins", "datetime", "persistent", "renku", "zc", "zope"]:
        raise TypeError(f"Objects of type '{type_name}' are not allowed")

    module = __import__(module_name)

    return get_attribute(module, components[1:])


def get_attribute(object, name: Union[List[str], str]):
    """Return an attribute of an object."""
    components = name.split(".") if isinstance(name, str) else name

    for component in components:
        object = getattr(object, component)

    return object


class Persistent(persistent.Persistent):
    """Base Persistent class for renku classes.

    Subclasses are assumed to be immutable once persisted to the database. If a class shouldn't be immutable then
    subclass it directly from persistent.Persistent.
    """

    _v_immutable = False

    def reassign_oid(self):
        """Reassign oid (after assigning a new identifier for example)."""
        if self._p_jar is not None:
            self._p_jar.remove_from_cache(self)

        self._p_oid = None
        self._p_oid = Database.generate_oid(self)

    @property
    def immutable(self):
        """Return if object is immutable."""
        return self._v_immutable

    def freeze(self):
        """Set immutable property."""
        self._v_immutable = True

    def __setattr__(self, key, value):
        if self._v_immutable and key != "__weakref__" and not key.startswith("_p_") and not key.startswith("_v_"):
            raise RuntimeError(f"Cannot modify immutable object {self}.{key}")

        super().__setattr__(key, value)


class Database:
    """The Metadata Object Database.

    This class is equivalent to a persistent.DataManager and implements persistent.interfaces.IPersistentDataManager
    interface.
    """

    ROOT_OID = "root"

    def __init__(self, storage):
        self._storage: Storage = storage
        self._cache = Cache()
        # The pre-cache is used by get to avoid infinite loops when objects load their state
        self._pre_cache: Dict[OID_TYPE, persistent.Persistent] = {}
        # Objects added explicitly by add() or when serializing other objects. After commit they are moved to _cache.
        self._objects_to_commit: Dict[OID_TYPE, persistent.Persistent] = {}
        self._reader: ObjectReader = ObjectReader(database=self)
        self._writer: ObjectWriter = ObjectWriter(database=self)
        self._root: Optional[OOBTree] = None

        self._initialize_root()

    @classmethod
    def from_path(cls, path: Union[Path, str]) -> "Database":
        """Create a Storage and Database using the given path."""
        storage = Storage(path)
        return Database(storage=storage)

    @staticmethod
    def generate_oid(object: persistent.Persistent) -> OID_TYPE:
        """Generate oid for a persistent.Persistent object based on its id."""
        oid = getattr(object, "_p_oid")
        if oid:
            assert isinstance(oid, OID_TYPE)
            return oid

        id: str = getattr(object, "id", None) or getattr(object, "_id", None)
        if id:
            return Database.hash_id(id)

        return Database.new_oid()

    @staticmethod
    def hash_id(id: str) -> OID_TYPE:
        """Return oid from id."""
        return hashlib.sha3_256(id.encode("utf-8")).hexdigest()

    @staticmethod
    def new_oid():
        """Generate a random oid."""
        return f"{uuid4().hex}{uuid4().hex}"

    @staticmethod
    def _get_filename_from_oid(oid: OID_TYPE) -> str:
        return oid.lower()

    def __getitem__(self, key) -> "Index":
        return self._root[key]

    def clear(self):
        """Remove all objects and clear all caches. Objects won't be deleted in the storage."""
        self._cache.clear()
        self._pre_cache.clear()
        self._objects_to_commit.clear()
        # NOTE: Clear root at the end because it will be added to _objects_to_commit when `register` is called.
        self._root.clear()

    def _initialize_root(self):
        """Initialize root object."""
        if not self._root:
            try:
                self._root = self.get(Database.ROOT_OID)
            except errors.ObjectNotFoundError:
                self._root = OOBTree()
                self._root._p_oid = Database.ROOT_OID
                self.register(self._root)

    def add_index(self, name: str, object_type: type, attribute: str = None, key_type: type = None) -> "Index":
        """Add an index."""
        assert name not in self._root, f"Index or object already exists: '{name}'"

        index = Index(name=name, object_type=object_type, attribute=attribute, key_type=key_type)
        index._p_jar = self

        self._root[name] = index

        return index

    def add_root_object(self, name: str, obj: Persistent):
        """Add an object to the DB root."""
        assert name not in self._root, f"Index or object already exists: '{name}'"

        obj._p_jar = self
        obj._p_oid = name

        self._root[name] = obj

    def add(self, object: persistent.Persistent, oid: OID_TYPE = None):
        """Add a new object to the database.

        NOTE: Normally, we add objects to indexes but this method adds objects directly to Dataset's root. Use it only
        for singleton objects that have no Index defined for them (e.g. Project).
        """
        assert not oid or isinstance(oid, OID_TYPE), f"Invalid oid type: '{type(oid)}'"
        object._p_oid = oid

        self.register(object)

    def register(self, object: persistent.Persistent):
        """Register a persistent.Persistent object to be stored.

        NOTE: When a persistent.Persistent object is changed it calls this method.
        """
        assert isinstance(object, persistent.Persistent), f"Cannot add non-Persistent object: '{object}'"

        if object._p_oid is None:
            object._p_oid = self.generate_oid(object)
        elif isinstance(object, Persistent):
            # NOTE: A safety-net to make sure that all objects have correct p_oid
            id = getattr(object, "id")
            expected_oid = Database.hash_id(id)
            actual_oid = object._p_oid
            assert actual_oid == expected_oid, f"Object has wrong oid: {actual_oid} != {expected_oid}"

        object._p_jar = self
        # object._p_serial = NEW
        self._objects_to_commit[object._p_oid] = object

    def get(self, oid: OID_TYPE) -> persistent.Persistent:
        """Get the object by oid."""
        if oid != Database.ROOT_OID and oid in self._root:  # NOTE: Avoid looping if getting "root"
            return self._root[oid]
        object = self.get_cached(oid)
        if object is not None:
            return object

        data = self._storage.load(filename=self._get_filename_from_oid(oid))
        object = self._reader.deserialize(data)
        object._p_changed = 0
        object._p_serial = PERSISTED
        if isinstance(object, Persistent):
            object.freeze()

        # NOTE: Avoid infinite loop if object tries to load its state before it is added to the cache
        self._pre_cache[oid] = object
        self._cache[oid] = object
        self._pre_cache.pop(oid)

        return object

    def get_by_id(self, id: str) -> persistent.Persistent:
        """Return an object by its id."""
        oid = Database.hash_id(id)
        return self.get(oid)

    def get_cached(self, oid: OID_TYPE) -> Optional[persistent.Persistent]:
        """Return an object if it is in the cache or will be committed."""
        object = self._cache.get(oid)
        if object is not None:
            return object

        object = self._pre_cache.get(oid)
        if object is not None:
            return object

        object = self._objects_to_commit.get(oid)
        if object is not None:
            return object

    def remove_root_object(self, name: str) -> None:
        """Remove a root object from the database."""
        assert name in self._root, f"Index or object doesn't exist in root: '{name}'"

        obj = self.get(name)
        self.remove_from_cache(obj)

        del self._root[name]

    def new_ghost(self, oid: OID_TYPE, object: persistent.Persistent):
        """Create a new ghost object."""
        object._p_jar = self
        self._cache.new_ghost(oid, object)

    def setstate(self, object: persistent.Persistent):
        """Load the state for a ghost object."""
        data = self._storage.load(filename=self._get_filename_from_oid(object._p_oid))
        self._reader.set_ghost_state(object, data)
        object._p_serial = PERSISTED
        if isinstance(object, Persistent):
            object.freeze()

    def commit(self):
        """Commit modified and new objects."""
        while self._objects_to_commit:
            _, object = self._objects_to_commit.popitem()
            if object._p_changed or object._p_serial == NEW:
                self._store_object(object)

    def _store_object(self, object: persistent.Persistent):
        data = self._writer.serialize(object)
        # TODO: Set compress based on the object type in the final version
        compress = False  # if isinstance(object, (OOBTree, OOBucket, Project, Index)) else True
        self._storage.store(filename=self._get_filename_from_oid(object._p_oid), data=data, compress=compress)

        self._cache[object._p_oid] = object

        object._p_changed = 0  # NOTE: transition from changed to up-to-date
        object._p_serial = PERSISTED

    def remove_from_cache(self, object: persistent.Persistent):
        """Remove an object from cache."""
        oid = object._p_oid

        def remove_from(cache):
            existing_entry = cache.get(oid)
            if existing_entry is object:
                cache.pop(oid)

        remove_from(self._cache)
        remove_from(self._pre_cache)
        remove_from(self._objects_to_commit)

    def readCurrent(self, object):
        """We don't use this method but some Persistent logic require its existence."""
        assert object._p_jar is self
        assert object._p_oid is not None

    def oldstate(self, object, tid):
        """See persistent.interfaces.IPersistentDataManager::oldstate."""
        raise NotImplementedError


@implementer(IPickleCache)
class Cache:
    """Database Cache."""

    def __init__(self):
        self._entries = {}

    def __len__(self):
        return len(self._entries)

    def __getitem__(self, oid):
        assert isinstance(oid, OID_TYPE), f"Invalid oid type: '{type(oid)}'"
        return self._entries[oid]

    def __setitem__(self, oid, object):
        assert isinstance(object, persistent.Persistent), f"Cannot cache non-Persistent objects: '{object}'"
        assert isinstance(oid, OID_TYPE), f"Invalid oid type: '{type(oid)}'"

        assert object._p_jar is not None, "Cached object jar missing"
        assert oid == object._p_oid, f"Cache key does not match oid: {oid} != {object._p_oid}"

        if oid in self._entries:
            existing_data = self.get(oid)
            if existing_data is not object:
                raise ValueError(f"The same oid exists: {existing_data} != {object}")

        self._entries[oid] = object

    def __delitem__(self, oid):
        assert isinstance(oid, OID_TYPE), f"Invalid oid type: '{type(oid)}'"
        self._entries.pop(oid)

    def clear(self):
        """Remove all entries."""
        self._entries.clear()

    def pop(self, oid, default=MARKER):
        """Remove and return an object."""
        return self._entries.pop(oid) if default is MARKER else self._entries.pop(oid, default)

    def get(self, oid, default=None):
        """See IPickleCache."""
        assert isinstance(oid, OID_TYPE), f"Invalid oid type: '{type(oid)}'"
        return self._entries.get(oid, default)

    def new_ghost(self, oid, object):
        """See IPickleCache."""
        assert object._p_oid is None, f"Object already has an oid: {object}"
        assert object._p_jar is not None, f"Object does not have a jar: {object}"
        assert oid not in self._entries, f"Duplicate oid: {oid}"

        object._p_oid = oid
        if object._p_state != GHOST:
            object._p_invalidate()

        self[oid] = object


class Index(persistent.Persistent):
    """Database index."""

    def __init__(self, *, name: str, object_type, attribute: Optional[str], key_type=None):
        """
        Create an index where keys are extracted using `attribute` from an object or a key.

        @param name: Index's name
        @param object_type: Type of objects that the index points to
        @param attribute: Name of an attribute to be used to automatically generate a key (e.g. `entity.path`)
        @param key_type: Type of keys. If not None then a key must be provided when updating the index
        """
        assert name == name.lower(), f"Index name must be all lowercase: '{name}'."

        super().__init__()

        self._p_oid = f"{name}-index"
        self._name: str = name
        self._object_type = object_type
        self._key_type = key_type
        self._attribute: Optional[str] = attribute
        self._entries: OOBTree = OOBTree()
        self._entries._p_oid = name

    def __len__(self):
        return len(self._entries)

    def __contains__(self, key):
        return key in self._entries

    def __getitem__(self, key):
        return self._entries[key]

    def __setitem__(self, key, value):
        # NOTE: if Index is using a key object then we cannot check if key is valid. It's safer to use `add` method
        # instead of setting values directly.
        self._verify_and_get_key(object=value, key_object=None, key=key, missing_key_object_ok=True)

        self._entries[key] = value

    def __getstate__(self):
        return {
            "name": self._name,
            "object_type": get_type_name(self._object_type),
            "key_type": get_type_name(self._key_type),
            "attribute": self._attribute,
            "entries": self._entries,
        }

    def __setstate__(self, data):
        self._name = data.pop("name")
        self._object_type = get_class(data.pop("object_type"))
        self._key_type = get_class(data.pop("key_type"))
        self._attribute = data.pop("attribute")
        self._entries = data.pop("entries")

    @property
    def name(self) -> str:
        """Return Index's name."""
        return self._name

    @property
    def object_type(self) -> type:
        """Return Index's object_type."""
        return self._object_type

    def get(self, key, default=None):
        """Return an entry based on its key."""
        return self._entries.get(key, default)

    def pop(self, key, default=MARKER):
        """Remove and return an object."""
        if not key:
            return
        return self._entries.pop(key) if default is MARKER else self._entries.pop(key, default)

    def keys(self):
        """Return an iterator of keys."""
        return self._entries.keys()

    def values(self):
        """Return an iterator of values."""
        return self._entries.values()

    def items(self):
        """Return an iterator of keys and values."""
        return self._entries.items()

    def add(self, object: persistent.Persistent, *, key: Optional[str] = None, key_object=None, verify=True):
        """Update index with object.

        If `Index._attribute` is not None then key is automatically generated.
        Key is extracted from `key_object` if it is not None; otherwise, it's extracted from `object`.
        """
        assert isinstance(object, self._object_type), f"Cannot add objects of type '{type(object)}'"
        key = self._verify_and_get_key(
            object=object, key_object=key_object, key=key, missing_key_object_ok=False, verify=verify
        )
        self._entries[key] = object

    def generate_key(self, object: persistent.Persistent, *, key_object=None):
        """Return index key for an object.

        Key is extracted from `key_object` if it is not None; otherwise, it's extracted from `object`.
        """
        return self._verify_and_get_key(object=object, key_object=key_object, key=None, missing_key_object_ok=False)

    def _verify_and_get_key(
        self, *, object: persistent.Persistent, key_object, key, missing_key_object_ok, verify=True
    ):
        if self._key_type:
            if not missing_key_object_ok:
                assert isinstance(key_object, self._key_type), f"Invalid key type: {type(key_object)} for '{self.name}'"
        else:
            assert key_object is None, f"Index '{self.name}' does not accept 'key_object'"

        if self._attribute:
            key_object = key_object or object
            correct_key = get_attribute(key_object, self._attribute)
            if key is not None:
                if verify:
                    assert key == correct_key, f"Incorrect key for index '{self.name}': '{key}' != '{correct_key}'"
                else:
                    correct_key = key
        else:
            assert key is not None, "No key is provided"
            correct_key = key

        return correct_key


class Storage:
    """Store Persistent objects on the disk."""

    OID_FILENAME_LENGTH = 64

    def __init__(self, path: Union[Path, str]):
        self.path = Path(path)

    def store(self, filename: str, data: Union[Dict, List], compress=False):
        """Store object."""
        assert isinstance(filename, str)

        is_oid_path = len(filename) == Storage.OID_FILENAME_LENGTH
        if is_oid_path:
            path = self.path / filename[0:2] / filename[2:4] / filename
            path.parent.mkdir(parents=True, exist_ok=True)
        else:
            path = self.path / filename
            self.path.mkdir(parents=True, exist_ok=True)

        open_func = gzip.open if compress else open

        with open_func(path, "wt") as file:
            json.dump(data, file, ensure_ascii=False, sort_keys=True, indent=2)

    def load(self, filename: str):
        """Load data for object with object id oid."""
        assert isinstance(filename, str)

        is_oid_path = len(filename) == Storage.OID_FILENAME_LENGTH
        if is_oid_path:
            path = self.path / filename[0:2] / filename[2:4] / filename
        else:
            path = self.path / filename

        if not path.exists():
            raise errors.ObjectNotFoundError(filename)

        try:
            with gzip.open(path) as file:
                data = json.load(file)
        except OSError:  # Not a gzip file
            with open(path) as file:
                data = json.load(file)

        return data


class ObjectWriter:
    """Serialize objects for storage in storage."""

    def __init__(self, database: Database):
        self._database: Database = database

    def serialize(self, object: persistent.Persistent):
        """Convert an object to JSON."""
        assert isinstance(object, persistent.Persistent), f"Cannot serialize object of type '{type(object)}': {object}"
        assert object._p_oid, f"Object does not have an oid: '{object}'"
        assert object._p_jar is not None, f"Object is not associated with a Database: '{object}'"

        self._serialization_cache = {}
        state = object.__getstate__()
        was_dict = isinstance(state, dict)
        data = self._serialize_helper(state)
        is_dict = isinstance(data, dict)

        if not is_dict or (is_dict and not was_dict):
            data = {"@renku_data_value": data}

        data["@renku_data_type"] = get_type_name(object)
        data["@renku_oid"] = object._p_oid

        return data

    def _serialize_helper(self, object):
        # TODO: Raise an error if an unsupported object is being serialized
        if object is None:
            return None
        elif isinstance(object, (int, float, str, bool)):
            return object
        elif isinstance(object, list):
            return [self._serialize_helper(value) for value in object]
        elif isinstance(object, dict):
            result = dict()
            items = sorted(object.items(), key=lambda x: x[0])
            for key, value in items:
                result[key] = self._serialize_helper(value)
            return result
        elif isinstance(object, Index):
            # NOTE: Index objects are not stored as references and are included in their parent object (i.e. root)
            state = object.__getstate__()
            state = self._serialize_helper(state)
            return {"@renku_data_type": get_type_name(object), "@renku_oid": object._p_oid, **state}
        elif isinstance(object, persistent.Persistent):
            if not object._p_oid:
                object._p_oid = Database.generate_oid(object)
            if object._p_state not in [GHOST, UPTODATE] or (object._p_state == UPTODATE and object._p_serial == NEW):
                self._database.register(object)
            return {"@renku_data_type": get_type_name(object), "@renku_oid": object._p_oid, "@renku_reference": True}
        elif isinstance(object, datetime.datetime):
            value = object.isoformat()
        elif isinstance(object, tuple):
            value = tuple(self._serialize_helper(value) for value in object)
        elif isinstance(object, (InterfaceClass)):
            # NOTE: Zope interfaces are weird, they're a class with type InterfaceClass, but need to be deserialized
            # as the class (without instantiation)
            return {"@renku_data_type": TYPE_TYPE, "@renku_data_value": f"{object.__module__}.{object.__name__}"}
        elif isinstance(object, type):
            # NOTE: We're storing a type, not an instance
            return {"@renku_data_type": TYPE_TYPE, "@renku_data_value": get_type_name(object)}
        elif isinstance(object, (FunctionType, BuiltinFunctionType)):
            name = object.__name__
            module = getattr(object, "__module__", None)
            return {"@renku_data_type": FUNCTION_TYPE, "@renku_data_value": f"{module}.{name}"}
        elif hasattr(object, "__getstate__"):
            if id(object) in self._serialization_cache:
                # NOTE: We already serialized this -> circular/repeat reference.
                return {"@renku_data_type": REFERENCE_TYPE, "@renku_data_value": self._serialization_cache[id(object)]}

            # NOTE: The reference used for circular reference is just the position in the serialization cache,
            # as the order is deterministic. So the order in which objects are encoutered is their id for referencing.
            self._serialization_cache[id(object)] = len(self._serialization_cache)

            value = object.__getstate__().copy()
            value = {k: v for k, v in value.items() if not k.startswith("_v_")}
            value = self._serialize_helper(value)
            assert not isinstance(value, dict) or "id" in value, f"Invalid object state: {value} for {object}"
        else:
            if id(object) in self._serialization_cache:
                # NOTE: We already serialized this -> circular/repeat reference
                return {"@renku_data_type": REFERENCE_TYPE, "@renku_data_value": self._serialization_cache[id(object)]}

            # NOTE: The reference used for circular reference is just the position in the serialization cache,
            # as the order is deterministic So the order in which objects are encoutered is their id for referencing.
            self._serialization_cache[id(object)] = len(self._serialization_cache)

            value = object.__dict__.copy()
            value = {k: v for k, v in value.items() if not k.startswith("_v_")}
            value = self._serialize_helper(value)

        return {"@renku_data_type": get_type_name(object), "@renku_data_value": value}


class ObjectReader:
    """Deserialize objects loaded from storage."""

    def __init__(self, database: Database):
        self._classes: Dict[str, type] = {}
        self._database = database

    def _get_class(self, type_name: str) -> type:
        cls = self._classes.get(type_name)
        if cls:
            return cls

        cls = get_class(type_name)

        self._classes[type_name] = cls
        return cls

    def set_ghost_state(self, object: persistent.Persistent, data: Dict):
        """Set state of a Persistent ghost object."""
        previous_cache = self._deserialization_cache
        self._deserialization_cache = []

        state = self._deserialize_helper(data, create=False)
        object.__setstate__(state)

        self._deserialization_cache = previous_cache

    def deserialize(self, data):
        """Convert JSON to Persistent object."""
        oid = data["@renku_oid"]

        self._deserialization_cache = []

        object = self._deserialize_helper(data)

        object._p_oid = oid
        object._p_jar = self._database

        return object

    def _deserialize_helper(self, data, create=True):
        if data is None:
            return None
        elif isinstance(data, (int, float, str, bool)):
            return data
        elif isinstance(data, list):
            return [self._deserialize_helper(value) for value in data]
        else:
            assert isinstance(data, dict), f"Data must be a dict: '{type(data)}'"

            if "@renku_data_type" not in data:  # NOTE: A normal dict value
                assert "@renku_oid" not in data
                items = sorted(data.items(), key=lambda x: x[0])
                for key, value in items:
                    data[key] = self._deserialize_helper(value)
                return data

            object_type = data.pop("@renku_data_type")
            if object_type in (TYPE_TYPE, FUNCTION_TYPE):
                # NOTE: if we stored a type (not instance), return the type
                return self._get_class(data["@renku_data_value"])
            elif object_type == REFERENCE_TYPE:
                # NOTE: we had a circular reference, we return the (not yet finalized) class here
                return self._deserialization_cache[data["@renku_data_value"]]

            cls = self._get_class(object_type)

            if issubclass(cls, datetime.datetime):
                assert create
                data = data["@renku_data_value"]
                return datetime.datetime.fromisoformat(data)
            elif issubclass(cls, tuple):
                data = data["@renku_data_value"]
                return tuple(self._deserialize_helper(value) for value in data)

            oid: str = data.pop("@renku_oid", None)
            if oid:
                assert isinstance(oid, str)

                if "@renku_reference" in data and data["@renku_reference"]:  # A reference
                    assert create, f"Cannot deserialize a reference without creating an instance {data}"
                    new_object = self._database.get_cached(oid)
                    if new_object is not None:
                        return new_object
                    assert issubclass(cls, persistent.Persistent)
                    new_object = cls.__new__(cls)
                    self._database.new_ghost(oid, new_object)
                    return new_object
                elif issubclass(cls, Index):
                    new_object = self._database.get_cached(oid)
                    if new_object:
                        return new_object
                    new_object = cls.__new__(cls)
                    new_object._p_oid = oid
                    self.set_ghost_state(new_object, data)
                    return new_object

            if "@renku_data_value" in data:
                data = data["@renku_data_value"]

            if not create:
                data = self._deserialize_helper(data)
                return data

            if issubclass(cls, persistent.Persistent):
                new_object = cls.__new__(cls)
                new_object._p_oid = oid
                self.set_ghost_state(new_object, data)
            else:
                new_object = cls.__new__(cls)

                # NOTE: we deserialize in the same order as we serialized, so the two stacks here match
                self._deserialization_cache.append(new_object)

                data = self._deserialize_helper(data)
                assert isinstance(data, dict)

                for name, value in data.items():
                    object.__setattr__(new_object, name, value)

                if issubclass(cls, Immutable):
                    new_object = cls.make_instance(new_object)

            return new_object



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client for handling a local repository."""

import hashlib
import json
import os
import shutil
import subprocess
from contextlib import contextmanager
from subprocess import check_output

import attr
import filelock
from jinja2 import Template
from werkzeug.utils import cached_property

from renku.core.compat import Path
from renku.core.management.command_builder import inject
from renku.core.management.config import RENKU_HOME
from renku.core.management.interface.database_gateway import IDatabaseGateway
from renku.core.management.interface.project_gateway import IProjectGateway
from renku.core.models.enums import ConfigFilter
from renku.core.models.project import Project
from renku.core.utils import communication

from .git import GitCore

DEFAULT_DATA_DIR = "data"

INIT_APPEND_FILES = [".gitignore"]
INIT_KEEP_FILES = ["readme.md", "readme.rst"]


def default_path(path="."):
    """Return default repository path."""
    from git import InvalidGitRepositoryError

    from renku.core.commands.git import get_git_home

    try:
        return get_git_home(path=path)
    except InvalidGitRepositoryError:
        return path


def path_converter(path):
    """Converter for path in PathMixin."""
    return Path(path).resolve()


@attr.s
class PathMixin:
    """Define a default path attribute."""

    path = attr.ib(default=default_path, converter=path_converter)

    @path.validator
    def _check_path(self, _, value):
        """Check the path exists and it is a directory."""
        if not (value.exists() and value.is_dir()):
            raise ValueError("Define an existing directory.")


@attr.s
class RepositoryApiMixin(GitCore):
    """Client for handling a local repository."""

    renku_home = attr.ib(default=RENKU_HOME)
    """Define a name of the Renku folder (default: ``.renku``)."""

    renku_path = attr.ib(init=False)
    """Store a ``Path`` instance of the Renku folder."""

    parent = attr.ib(default=None)
    """Store a pointer to the parent repository."""

    data_dir = attr.ib(
        default=DEFAULT_DATA_DIR, kw_only=True, converter=lambda value: str(value) if value else DEFAULT_DATA_DIR
    )
    """Define a name of the folder for storing datasets."""

    LOCK_SUFFIX = ".lock"
    """Default suffix for Renku lock file."""

    DATABASE_PATH: str = "metadata"
    """Directory for metadata storage."""

    DOCKERFILE = "Dockerfile"
    """Name of the Dockerfile in the repo."""

    TEMPLATE_CHECKSUMS = "template_checksums.json"

    RENKU_PROTECTED_PATHS = [
        ".dockerignore",
        ".git",
        ".git/*",
        ".gitattributes",
        ".gitignore",
        ".gitlab-ci.yml",
        ".renku",
        ".renku/*",
        ".renkulfsignore",
        "Dockerfile",
        "environment.yml",
        "requirements.txt",
    ]

    _remote_cache = attr.ib(factory=dict)

    def __attrs_post_init__(self):
        """Initialize computed attributes."""
        #: Configure Renku path.
        path = Path(self.renku_home)
        if not path.is_absolute():
            path = self.path / path

        path.relative_to(path)
        self.renku_path = path

        data_dir = self.get_value("renku", self.DATA_DIR_CONFIG_KEY, config_filter=ConfigFilter.LOCAL_ONLY)
        self.data_dir = data_dir or self.data_dir

        self._subclients = {}

        self._project = None

        super().__attrs_post_init__()

        # initialize submodules
        if self.repo:
            try:
                check_output(["git", "submodule", "update", "--init", "--recursive"], cwd=str(self.path))
            except subprocess.CalledProcessError:
                pass

    @property
    def latest_agent(self):
        """Returns latest agent version used in the repository."""
        try:
            return self.project.agent_version
        except ValueError as e:
            if "Project name not set" in str(e):
                return None
            raise

    @property
    def lock(self):
        """Create a Renku config lock."""
        return filelock.FileLock(str(self.renku_path.with_suffix(self.LOCK_SUFFIX)), timeout=0)

    @property
    def docker_path(self):
        """Path to the Dockerfile."""
        return self.path / self.DOCKERFILE

    @property
    def template_checksums(self):
        """Return a ``Path`` instance to the template checksums file."""
        return self.renku_path / self.TEMPLATE_CHECKSUMS

    @property
    def database_path(self) -> Path:
        """Path to the metadata storage directory."""
        return self.renku_path / self.DATABASE_PATH

    @property
    @inject.autoparams()
    def project(self, project_gateway: IProjectGateway):
        """Return the Project instance."""
        if self._project is None:
            self._project = project_gateway.get_project()

        return self._project

    @property
    def remote(self, remote_name="origin"):
        """Return host, owner and name of the remote if it exists."""
        from renku.core.models.git import GitURL

        original_remote_name = remote_name

        if original_remote_name in self._remote_cache:
            return self._remote_cache[original_remote_name]

        host = owner = name = None
        try:
            remote_branch = self.repo.head.reference.tracking_branch()
            if remote_branch is not None:
                remote_name = remote_branch.remote_name
        except TypeError:
            pass

        try:
            url = GitURL.parse(self.repo.remotes[remote_name].url)

            # Remove gitlab. unless running on gitlab.com.
            hostname_parts = url.hostname.split(".")
            if len(hostname_parts) > 2 and hostname_parts[0] == "gitlab":
                hostname_parts = hostname_parts[1:]
            url = attr.evolve(url, hostname=".".join(hostname_parts))
        except IndexError:
            url = None

        if url:
            host = url.hostname
            owner = url.owner
            name = url.name

        remote = {"host": host, "owner": owner, "name": name}
        self._remote_cache[original_remote_name] = remote

        return remote

    def is_project_set(self):
        """Return if project is set for the client."""
        return self._project is not None

    def find_previous_commit(self, paths, revision="HEAD", return_first=False, full=False):
        """Return a previous commit for a given path starting from ``revision``.

        :param revision: revision to start from, defaults to ``HEAD``
        :param return_first: show the first commit in the history
        :param full: return full history
        :raises KeyError: if path is not present in the given commit
        """
        kwargs = {}

        if full:
            kwargs["full_history"] = True

        if return_first:
            file_commits = list(self.repo.iter_commits(revision, paths=paths, **kwargs))
        else:
            file_commits = list(self.repo.iter_commits(revision, paths=paths, max_count=1, **kwargs))

        if not file_commits:
            raise KeyError("Could not find a file {0} in range {1}".format(paths, revision))

        return file_commits[-1 if return_first else 0]

    @cached_property
    def submodules(self):
        """Return list of submodules it belongs to."""
        if self.parent:
            client, submodule = self.parent
            return client.submodules + [submodule.name]
        return []

    def subclients(self, parent_commit):
        """Return mapping from submodule to client."""
        if parent_commit in self._subclients:
            return self._subclients[parent_commit]

        try:
            from git import Submodule

            submodules = [submodule for submodule in Submodule.iter_items(self.repo, parent_commit=parent_commit)]
        except (RuntimeError, ValueError):
            # There are no submodules associated with the given commit.
            submodules = []

        subclients = {}
        for submodule in submodules:
            subpath = (self.path / submodule.path).resolve()
            is_renku = subpath / Path(self.renku_home)

            if subpath.exists() and is_renku.exists():
                subclients[submodule] = self.__class__(path=subpath, parent=(self, submodule))

        return subclients

    def resolve_in_submodules(self, commit, path):
        """Resolve filename in submodules."""
        original_path = self.path / path
        in_vendor = str(path).startswith(".renku/vendors")

        if original_path.is_symlink() or in_vendor:
            original_path = Path(os.path.realpath(os.path.abspath(str(original_path))))

            for submodule, subclient in self.subclients(commit).items():
                if (Path(submodule.path) / Path(".git")).exists():

                    try:
                        subpath = original_path.relative_to(subclient.path)
                        return (
                            subclient,
                            subclient.find_previous_commit(subpath, revision=submodule.hexsha),
                            subpath,
                        )
                    except ValueError:
                        pass

        return self, commit, path

    @contextmanager
    @inject.autoparams()
    def with_metadata(
        self,
        project_gateway: IProjectGateway,
        database_gateway: IDatabaseGateway,
        read_only=False,
        name=None,
        description=None,
        custom_metadata=None,
    ):
        """Yield an editable metadata object."""

        try:
            project = project_gateway.get_project()
        except ValueError:
            project = Project.from_client(
                name=name, description=description, custom_metadata=custom_metadata, client=self
            )

        yield project

        if not read_only:
            project_gateway.update_project(project)
            database_gateway.commit()

    def has_graph_files(self):
        """Return true if database exists."""
        return self.database_path.exists() and any(
            f for f in self.database_path.iterdir() if f != self.database_path / "root"
        )

    def init_repository(self, force=False, user=None, initial_branch=None):
        """Initialize an empty Renku repository."""
        from git import Repo

        from renku.core.models.provenance.agent import Person

        # initialize repo and set user data
        kwargs = {}

        if initial_branch:
            kwargs["initial-branch"] = initial_branch

        path = self.path.absolute()
        self.repo = Repo.init(str(path), **kwargs)
        if user:
            config_writer = self.repo.config_writer()
            for key, value in user.items():
                config_writer.set_value("user", key, value)
            config_writer.release()

        # verify if author information is available
        Person.from_git(self.repo)

    def get_template_files(self, template_path, metadata):
        """Gets paths in a rendered renku template."""
        for file in template_path.glob("**/*"):
            rel_path = file.relative_to(template_path)
            destination = self.path / rel_path

            destination = Path(Template(str(destination)).render(metadata))
            yield destination.relative_to(self.path)

    def import_from_template(self, template_path, metadata, force=False):
        """Render template files from a template directory."""
        checksums = {}
        for file in sorted(template_path.glob("**/*")):
            rel_path = file.relative_to(template_path)
            destination = self.path / rel_path

            try:
                # TODO: notify about the expected variables - code stub:
                # with file.open() as fr:
                #     file_content = fr.read()
                #     # look for the required keys
                #     env = Environment()
                #     parsed = env.parse(file_content)
                #     variables = meta.find_undeclared_variables(parsed)

                # parse file and process it
                template = Template(file.read_text())
                rendered_content = template.render(metadata)
                # NOTE: the path could contain template variables, we need to template it
                destination = Path(Template(str(destination)).render(metadata))
                templated_rel_path = destination.relative_to(self.path)

                if destination.exists() and str(templated_rel_path).lower() in INIT_APPEND_FILES:
                    communication.echo(f"Appending to file {templated_rel_path} ...")
                    destination.write_text(destination.read_text() + "\n" + rendered_content)
                elif not destination.exists() or str(templated_rel_path).lower() not in INIT_KEEP_FILES:
                    if destination.exists():
                        communication.echo(f"Overwriting file {templated_rel_path} ...")
                    else:
                        communication.echo(f"Initializing file {templated_rel_path} ...")

                    destination.write_text(rendered_content)

                checksums[str(rel_path)] = self._content_hash(destination)
            except IsADirectoryError:
                destination.mkdir(parents=True, exist_ok=True)
            except TypeError:
                shutil.copy(file, destination)

        self.template_checksums.parent.mkdir(parents=True, exist_ok=True)

        with open(self.template_checksums, "w") as checksum_file:
            json.dump(checksums, checksum_file)

    def _content_hash(self, path):
        """Calculate the sha256 hash of a file."""
        sha256_hash = hashlib.sha256()
        with open(str(path), "rb") as f:
            for byte_block in iter(lambda: f.read(4096), b""):
                sha256_hash.update(byte_block)
        return sha256_hash.hexdigest()


DATABASE_METADATA_PATH = [
    Path(RENKU_HOME) / RepositoryApiMixin.DATABASE_PATH,
]



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test metadata Database."""

import pytest
from persistent import GHOST, UPTODATE
from persistent.list import PersistentList
from persistent.mapping import PersistentMapping

from renku.core.metadata.database import PERSISTED, Database
from renku.core.models.entity import Entity
from renku.core.models.provenance.activity import Activity, Association, Usage
from renku.core.models.workflow.plan import Plan


def test_database_add(database):
    """Test adding an object to an index."""
    database, storage = database

    id = "/activities/42"
    activity = Activity(id=id)
    index = database.get("activities")
    index.add(activity)
    database.commit()

    root_objects = ["root", "activities", "plans"]
    for filename in root_objects:
        assert storage.exists(filename)

    oid = Database.hash_id(id)
    assert storage.exists(oid)


def test_database_add_using_set_item(database):
    """Test adding an object to the database using __setitem__."""
    database, storage = database

    id = "/activities/42"
    activity = Activity(id=id)
    database["activities"][id] = activity

    assert {id} == set(database["activities"].keys())
    assert {activity} == set(database["activities"].values())


def test_database_index_with_no_automatic_key(database):
    """Test indexes with no automatic key attribute."""
    database, storage = database
    index = database.add_index(name="manual", object_type=Activity)

    id = "/activities/42"
    activity = Activity(id=id)
    index.add(activity, key=id)

    database.commit()

    new_database = Database(storage=storage)
    activity = new_database["manual"][id]

    assert id == activity.id

    oid = Database.hash_id(id)
    assert storage.exists(oid)


def test_database_add_with_incorrect_key(database):
    """Test adding an object to the database using __setitem__ with an incorrect key should fail."""
    database, storage = database

    id = "/activities/42"
    activity_1 = Activity(id=id)

    with pytest.raises(AssertionError) as e:
        database["activities"]["incorrect-key"] = activity_1

    assert "Incorrect key for index 'activities': 'incorrect-key' != '/activities/42'" in str(e)


def test_database_add_fails_when_no_key_and_no_automatic_key(database):
    """Test adding to an index with no automatic key fails if no key is provided."""
    database, storage = database
    index = database.add_index(name="manual", object_type=Activity)

    activity = Activity(id="/activities/42")

    with pytest.raises(AssertionError) as e:
        index.add(activity)

    assert "No key is provided" in str(e)


def test_database_no_file_created_if_not_committed(database):
    """Test adding an object to a database does not create a file before commit."""
    database, storage = database
    database.commit()

    assert storage.exists("root")

    id = "/activities/42"
    activity = Activity(id=id)
    database.get("activities").add(activity)

    oid = Database.hash_id(id)
    assert not storage.exists(oid)


def test_database_update_required_object_only(database):
    """Test adding an object to the database does not cause an update to all other objects."""
    database, storage = database

    index = database.get("activities")

    id_1 = "/activities/42"
    activity_1 = Activity(id=id_1)
    index.add(activity_1)
    database.commit()
    oid_1 = Database.hash_id(id_1)
    modification_time_before = storage.get_modification_date(oid_1)

    id_2 = "/activities/43"
    activity_2 = Activity(id=id_2)
    index.add(activity_2)
    database.commit()

    modification_time_after = storage.get_modification_date(oid_1)

    assert modification_time_before == modification_time_after


def test_database_update_required_root_objects_only(database):
    """Test adding an object to an index does not cause an update to other indexes."""
    database, storage = database

    database.commit()

    entity_modification_time_before = storage.get_modification_date("plans")
    activity_modification_time_before = storage.get_modification_date("activities")

    activity = Activity(id="/activities/42")
    database.get("activities").add(activity)
    database.commit()

    entity_modification_time_after = storage.get_modification_date("plans")
    activity_modification_time_after = storage.get_modification_date("activities")

    assert entity_modification_time_before == entity_modification_time_after
    assert activity_modification_time_before != activity_modification_time_after


def test_database_add_non_persistent(database):
    """Test adding a non-Persistent object to the database raises an error."""
    database, _ = database

    class Dummy:
        id = 42

    with pytest.raises(AssertionError) as e:
        object = Dummy()
        database.get("activities").add(object)

    assert "Cannot add objects of type" in str(e)


def test_database_loads_only_required_objects(database):
    """Test loading an object does not load its Persistent members."""
    database, storage = database

    plan = Plan(id="/plan/9")
    association = Association(id="association", plan=plan)
    id = "/activities/42"
    activity = Activity(id=id, association=association)
    database.get("activities").add(activity)
    database.commit()

    new_database = Database(storage=storage)
    oid = Database.hash_id(id)
    activity = new_database.get(oid)

    # Access a field to make sure that activity is loaded
    _ = activity.id

    assert UPTODATE == activity._p_state, activity._p_status
    assert PERSISTED == activity._p_serial
    assert GHOST == activity.association.plan._p_state

    assert UPTODATE == new_database["plans"]._p_state
    assert UPTODATE == new_database["activities"]._p_state


def test_database_load_multiple(database):
    """Test loading an object from multiple indexes returns the same object."""
    database, storage = database
    database.add_index(name="associations", object_type=Activity, attribute="association.id")

    plan = Plan(id="/plan/9")
    association = Association(id="/association/42", plan=plan)
    id = "/activities/42"
    activity = Activity(id=id, association=association)
    database.get("activities").add(activity)
    database.get("associations").add(activity)
    database.commit()

    new_database = Database(storage=storage)
    oid = Database.hash_id(id)
    activity_1 = new_database.get(oid)
    activity_2 = new_database.get("activities").get(id)
    activity_3 = new_database.get("associations").get("/association/42")

    assert activity_1 is activity_2
    assert activity_2 is activity_3


def test_database_index_update(database):
    """Test adding objects with the same key, updates the Index."""
    database, storage = database

    index_name = "plan-names"
    database.add_index(name=index_name, object_type=Plan, attribute="name")

    name = "same-name"

    plan_1 = Plan(id="/plans/42", name=name, description="old")
    database.get(index_name).add(plan_1)
    plan_2 = Plan(id="/plans/43", name=name, description="new")
    database.get(index_name).add(plan_2)
    assert plan_2 is database.get(index_name).get(name)

    database.commit()

    plan_3 = Plan(id="/plans/44", name=name, description="newer")
    database.get(index_name).add(plan_3)
    database.commit()

    new_database = Database(storage=storage)
    plans = new_database.get(index_name)
    plan = plans.get(name)

    assert "newer" == plan.description


def test_database_add_duplicate_index(database):
    """Test cannot add an index with the same name."""
    database, _ = database

    same_name = "plans"

    with pytest.raises(AssertionError) as e:
        database.add_index(name=same_name, object_type=Plan, attribute="name")

    assert "Index or object already exists: 'plans'" in str(e)


def test_database_index_different_key_type(database):
    """Test adding an Index with a different key type."""
    database, storage = database

    index_name = "usages"
    index = database.add_index(name=index_name, object_type=Activity, attribute="entity.path", key_type=Usage)

    entity = Entity(checksum="42", path="dummy/path")
    usage = Usage(entity=entity, id="/usages/42")

    activity = Activity(id="/activities/42", usages=[usage])
    database.get(index_name).add(activity, key_object=usage)
    database.commit()

    new_database = Database(storage=storage)
    usages = new_database[index_name]
    activity = usages.get("dummy/path")

    assert "/activities/42" == activity.id
    assert "42" == activity.usages[0].entity.checksum
    assert "dummy/path" == activity.usages[0].entity.path

    key = index.generate_key(activity, key_object=usage)

    assert activity is usages[key]


def test_database_wrong_index_key_type(database):
    """Test adding to an Index with a wrong key type."""
    database, _ = database

    index_name = "usages"
    database.add_index(name=index_name, object_type=Activity, attribute="id", key_type=Usage)

    activity = Activity(id="/activities/42")

    with pytest.raises(AssertionError) as e:
        database.get(index_name).add(activity)

    assert "Invalid key type" in str(e)


def test_database_missing_attribute(database):
    """Test adding to an Index while object does not have the requires attribute."""
    database, _ = database

    index_name = "usages"
    database.add_index(name=index_name, object_type=Activity, attribute="missing.attribute")

    activity = Activity(id="/activities/42")

    with pytest.raises(AttributeError) as e:
        database.get(index_name).add(activity)

    assert "'Activity' object has no attribute 'missing'" in str(e)


def test_database_remove(database):
    """Test removing an object from an index."""
    database, storage = database

    id = "/activities/42"
    activity = Activity(id=id)
    database.get("activities").add(activity)
    database.commit()

    database = Database(storage=storage)
    database.get("activities").pop(id)
    database.commit()

    database = Database(storage=storage)
    activity = database.get("activities").get(id, None)

    assert activity is None
    # However, the file still exists in the storage
    oid = Database.hash_id(id)
    assert storage.exists(oid)


def test_database_remove_non_existing(database):
    """Test removing a non-existing object from an index."""
    database, storage = database

    with pytest.raises(KeyError):
        database.get("activities").pop("non-existing-key")

    object = database.get("activities").pop("non-existing-key", None)

    assert object is None


def test_database_persistent_collections(database):
    """Test using Persistent collections."""
    database, storage = database
    index_name = "collections"
    database.add_index(name=index_name, object_type=PersistentMapping)

    entity_checksum = "42"
    entity_path = "dummy/path"
    usage = Usage(entity=Entity(checksum=entity_checksum, path=entity_path), id="/usages/42")
    id_1 = "/activities/1"
    activity_1 = Activity(id=id_1, usages=[usage])
    id_2 = "/activities/2"
    activity_2 = Activity(id=id_2, usages=[usage])

    p_mapping = PersistentMapping()

    database[index_name][entity_path] = p_mapping

    p_list = PersistentList()
    p_mapping[entity_checksum] = p_list
    p_list.append(activity_1)
    p_list.append(activity_2)

    database.commit()

    new_database = Database(storage=storage)
    collections = new_database[index_name]

    id_3 = "/activities/3"
    activity_3 = Activity(id=id_3, usages=[usage])
    collections[entity_path][entity_checksum].append(activity_3)

    assert {id_1, id_2, id_3} == {activity.id for activity in collections[entity_path][entity_checksum]}


def test_database_immutable_object(database):
    """Test storage and retrieval of immutable objects."""
    database, storage = database
    index_name = "collections"
    database.add_index(name=index_name, object_type=PersistentMapping)

    usage = Usage(entity=Entity(checksum="42", path="dummy/path"), id="/usages/42")
    id_1 = "/activities/1"
    activity_1 = Activity(id=id_1, usages=[usage])
    id_2 = "/activities/2"
    activity_2 = Activity(id=id_2, usages=[usage])

    database.get("activities").add(activity_1)
    database.get("activities").add(activity_2)
    database.commit()

    database = Database(storage=storage)
    activity_1 = database["activities"][id_1]
    activity_2 = database["activities"][id_2]

    usage_1 = activity_1.usages[0]
    usage_2 = activity_1.usages[0]

    assert isinstance(usage_1, Usage)
    assert usage_1 is usage_2



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test ``workflow`` commands."""

import os
import tempfile
from pathlib import Path

import pytest
from cwl_utils import parser_v1_2 as cwlgen

from renku.cli import cli
from renku.core.metadata.database import Database
from renku.core.models.jsonld import write_yaml
from renku.core.plugins.provider import available_workflow_providers
from tests.utils import format_result_exception


def test_workflow_compose(runner, project, run_shell, client):
    """Test renku workflow compose."""
    # Run a shell command with pipe.
    output = run_shell('renku run --name run1 -- echo "a" > output1')

    # Assert expected empty stdout.
    assert b"" == output[0]
    # Assert not allocated stderr.
    assert output[1] is None

    # Run a shell command with pipe.
    output = run_shell("renku run --name run2 -- cp output1 output2")

    # Assert expected empty stdout.
    assert b"" == output[0]
    # Assert not allocated stderr.
    assert output[1] is None

    result = runner.invoke(
        cli,
        [
            "workflow",
            "compose",
            "--map",
            "input_str=@step1.@param1",
            "--map",
            "output_file=run2.@output1",
            "--link",
            "@step1.@output1=@step2.@input1",
            "--set",
            "input_str=b",
            "--set",
            "output_file=other_output.csv",
            "-p",
            "input_str=the input string for the workflow",
            "-p",
            "output_file=the final output file produced",
            "composite_workflow",
            "run1",
            "run2",
        ],
    )

    assert 0 == result.exit_code, format_result_exception(result)

    database = Database.from_path(client.database_path)

    composite_plan = database["plans-by-name"]["composite_workflow"]

    assert composite_plan

    assert len(composite_plan.plans) == 2
    assert len(composite_plan.mappings) == 2

    assert composite_plan.mappings[0].name == "input_str"
    assert composite_plan.mappings[0].default_value == "b"
    assert composite_plan.mappings[0].description == "the input string for the workflow"

    assert composite_plan.mappings[1].name == "output_file"
    assert composite_plan.mappings[1].default_value == "other_output.csv"
    assert composite_plan.mappings[1].description == "the final output file produced"


def test_workflow_show(runner, project, run_shell, client):
    """Test renku workflow show."""
    # Run a shell command with pipe.
    output = run_shell('renku run --name run1 --description "my workflow" --success-code 0 -- echo "a" > output1')

    # Assert expected empty stdout.
    assert b"" == output[0]
    # Assert not allocated stderr.
    assert output[1] is None

    # Run a shell command with pipe.
    output = run_shell("renku run --name run2 -- cp output1 output2")

    # Assert expected empty stdout.
    assert b"" == output[0]
    # Assert not allocated stderr.
    assert output[1] is None

    result = runner.invoke(cli, ["workflow", "show", "run1"])
    assert 0 == result.exit_code, format_result_exception(result)
    assert "run1" in result.output
    assert "output1" in result.output
    assert "my workflow" in result.output

    result = runner.invoke(
        cli,
        [
            "workflow",
            "compose",
            "--description",
            "My composite workflow",
            "--map",
            "input_str=@step1.@param1",
            "--map",
            "output_file=run2.@output1",
            "--link",
            "@step1.@output1=@step2.@input1",
            "--set",
            "input_str=b",
            "--set",
            "output_file=other_output.csv",
            "-p",
            "input_str=the input string for the workflow",
            "-p",
            "output_file=the final output file produced",
            "composite_workflow",
            "run1",
            "run2",
        ],
    )

    assert 0 == result.exit_code, format_result_exception(result)

    result = runner.invoke(cli, ["workflow", "show", "composite_workflow"])

    assert 0 == result.exit_code, format_result_exception(result)
    assert "composite_workflow" in result.output
    assert "input_str" in result.output
    assert "output_file" in result.output
    assert "Links:" in result.output
    assert "Mappings:" in result.output
    assert "My composite workflow" in result.output


def test_workflow_remove_command(runner, project):
    """test workflow remove with builder."""
    workflow_name = "test_workflow"

    result = runner.invoke(cli, ["workflow", "remove", workflow_name])
    assert 2 == result.exit_code

    result = runner.invoke(cli, ["run", "--success-code", "0", "--no-output", "--name", workflow_name, "echo", "foo"])
    assert 0 == result.exit_code

    result = runner.invoke(cli, ["workflow", "remove", "--force", workflow_name])
    assert 0 == result.exit_code


def test_workflow_export_command(runner, project):
    """test workflow export with builder."""
    result = runner.invoke(cli, ["run", "--success-code", "0", "--no-output", "--name", "run1", "touch", "data.csv"])
    assert 0 == result.exit_code

    result = runner.invoke(cli, ["workflow", "export", "run1", "-o", "run1.cwl"])
    assert 0 == result.exit_code
    assert Path("run1.cwl").exists()

    workflow = cwlgen.load_document("run1.cwl")
    assert workflow.baseCommand[0] == "touch"
    assert len(workflow.inputs) == 3
    assert len(workflow.outputs) == 1


def test_workflow_edit(runner, client, run_shell):
    """Test naming of CWL tools and workflows."""

    def _get_plan_id(output):
        return output.split("\n")[0].split(":")[1].strip()

    workflow_name = "run1"
    result = runner.invoke(cli, ["run", "--name", workflow_name, "touch", "data.txt"])
    assert 0 == result.exit_code, format_result_exception(result)

    cmd = ["workflow", "edit", workflow_name, "--name", "first"]
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code, format_result_exception(result)

    database = Database.from_path(client.database_path)

    test_plan = database["plans-by-name"][workflow_name]
    first_plan = database["plans-by-name"]["first"]

    assert first_plan
    assert first_plan.name == "first"
    assert first_plan.derived_from == test_plan.id

    cmd = ["workflow", "edit", workflow_name, "--description", "Test workflow"]
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code, format_result_exception(result)

    database = Database.from_path(client.database_path)
    first_plan = database["plans"][_get_plan_id(result.stdout)]
    assert first_plan.description == "Test workflow"

    # edit parameter
    cmd = ["workflow", "edit", workflow_name, "--rename-param", "param1=param2"]
    result = runner.invoke(cli, cmd)
    assert 0 != result.exit_code, format_result_exception(result)

    cmd = ["workflow", "edit", workflow_name, "--set", "param1=0"]
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code, format_result_exception(result)
    edited_plan_id = _get_plan_id(result.output)

    database = Database.from_path(client.database_path)
    renamed_param_plan = database["plans"][_get_plan_id(result.output)]
    assert len(renamed_param_plan.parameters) > 0

    cmd = ["workflow", "edit", workflow_name, "--rename-param", "param1=param2"]
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code, format_result_exception(result)

    database = Database.from_path(client.database_path)
    renamed_param_plan = database["plans"][_get_plan_id(result.output)]
    parameter_names = list(map(lambda x: x.name, renamed_param_plan.parameters))
    assert len(parameter_names) > 0
    assert "param1" not in parameter_names
    assert "param2" in parameter_names
    param2_plan_id = _get_plan_id(result.stdout)

    # edit parameter description
    cmd = ["workflow", "edit", param2_plan_id, "--describe-param", "param1=Test"]
    result = runner.invoke(cli, cmd)
    assert 0 != result.exit_code, format_result_exception(result)

    cmd = ["workflow", "edit", param2_plan_id, "--describe-param", 'param2="Test parameter"']
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code, format_result_exception(result)

    database = Database.from_path(client.database_path)
    renamed_param_plan = database["plans"][_get_plan_id(result.output)]
    assert "Test parameter" == renamed_param_plan.parameters[0].description

    # edit parameter mapping
    cmd = ["run", "--name", "run2", "cp", "data.txt", "data2.txt"]
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code, format_result_exception(result)

    result = runner.invoke(
        cli,
        [
            "workflow",
            "compose",
            "--description",
            "My composite workflow",
            "--map",
            "input_str=@step1.@param1",
            "--map",
            "output_file=run2.@output1",
            "--link",
            "@step1.@output1=@step2.@input1",
            "--set",
            "input_str=b",
            "--set",
            "output_file=other_output.csv",
            "-p",
            "input_str=the input string for the workflow",
            "-p",
            "output_file=the final output file produced",
            "composite_workflow",
            edited_plan_id,
            "run2",
        ],
    )

    assert 0 == result.exit_code, format_result_exception(result)

    # remove mapping
    cmd = ["workflow", "edit", "composite_workflow", "--map", "output_file="]
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code, format_result_exception(result)

    database = Database.from_path(client.database_path)
    edited_composite_plan = database["plans"][_get_plan_id(result.output)]
    assert len(edited_composite_plan.mappings) == 1
    assert edited_composite_plan.mappings[0].mapped_parameters[0].name == "param1"


def test_workflow_show_outputs_with_directory(runner, client, run):
    """Output files in directory are not shown as separate outputs."""
    base_sh = ["bash", "-c", 'DIR="$0"; mkdir -p "$DIR"; ' 'for x in "$@"; do touch "$DIR/$x"; done']

    assert 0 == run(args=["run"] + base_sh + ["output", "foo", "bar"])
    assert (client.path / "output" / "foo").exists()
    assert (client.path / "output" / "bar").exists()

    cmd = ["workflow", "outputs"]
    result = runner.invoke(cli, cmd)
    assert 0 == result.exit_code, format_result_exception(result)
    assert {"output"} == set(result.output.strip().split("\n"))

    result = runner.invoke(cli, cmd + ["output"])
    assert 0 == result.exit_code, format_result_exception(result)
    assert {"output"} == set(result.output.strip().split("\n"))

    result = runner.invoke(cli, cmd + ["output/foo"])
    assert 0 == result.exit_code, format_result_exception(result)
    assert {"output"} == set(result.output.strip().split("\n"))

    result = runner.invoke(cli, cmd + ["output/foo", "output/bar"])
    assert 0 == result.exit_code, format_result_exception(result)
    assert {"output"} == set(result.output.strip().split("\n"))


@pytest.mark.parametrize("provider", available_workflow_providers())
@pytest.mark.parametrize("yaml", [False, True])
@pytest.mark.parametrize(
    "workflows, parameters",
    [
        ([("run", 'echo "a" > output1')], {}),
        ([("run", 'echo "a" > output1')], {"run": {"outputs": ["replaced"]}}),
        ([("run", 'echo "a" > output1')], {"run": {"parameters": ["foo"], "outputs": ["bar"]}}),
        (
            [("run1", "touch data.csv"), ("run2", "wc data.csv > output")],
            {"run1": {"outputs": ["foo"]}, "run2": {"inputs": ["foo"], "outputs": ["bar"]}},
        ),
    ],
)
def test_workflow_execute_command(runner, run_shell, project, capsys, client, provider, yaml, workflows, parameters):
    """test workflow execute."""

    for wf in workflows:
        output = run_shell(f"renku run --name {wf[0]} -- {wf[1]}")
        # Assert expected empty stdout.
        assert b"" == output[0]
        # Assert not allocated stderr.
        assert output[1] is None

    def _execute(args):
        with capsys.disabled():
            try:
                cli.main(
                    args=args,
                    prog_name=runner.get_default_prog_name(cli),
                )
            except SystemExit as e:
                assert e.code in {None, 0}

    if not parameters:
        for wf in workflows:
            execute_cmd = ["workflow", "execute", "-p", provider, wf[0]]
            _execute(execute_cmd)
    else:
        database = Database.from_path(client.database_path)
        for wf in workflows:
            if wf[0] in parameters:
                plan = database["plans-by-name"][wf[0]]
                execute_cmd = ["workflow", "execute", "-p", provider]

                overrides = dict()
                for k, values in parameters[wf[0]].items():
                    for i, v in enumerate(values):
                        overrides[getattr(plan, k)[i].name] = v

                if yaml:
                    fd, values_path = tempfile.mkstemp()
                    os.close(fd)
                    write_yaml(values_path, overrides)
                    execute_cmd += ["--values", values_path]
                else:
                    [execute_cmd.extend(["--set", f"{k}={v}"]) for k, v in overrides.items()]

                execute_cmd.append(wf[0])

                _execute(execute_cmd)

                # check whether parameters setting was effective
                if "outputs" in parameters[wf[0]]:
                    for o in parameters[wf[0]]["outputs"]:
                        assert Path(o).resolve().exists()



# -*- coding: utf-8 -*-
#
# Copyright 2018-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OS utility functions."""

import os
from pathlib import Path
from typing import List, Union

from renku.core import errors


def get_relative_path_to_cwd(path: Union[Path, str]) -> str:
    """Get a relative path to current working directory."""
    absolute_path = os.path.abspath(path)
    return os.path.relpath(absolute_path, os.getcwd())


def get_relative_paths(base: Union[Path, str], paths: List[Union[Path, str]]) -> List[str]:
    """Return a list of paths relative to a base path."""
    relative_paths = []

    for path in paths:
        try:
            # NOTE: Do not use os.path.realpath or Path.resolve() because they resolve symlinks
            absolute_path = os.path.abspath(os.path.join(base, path))
            relative_path = Path(absolute_path).relative_to(base)
        except ValueError:
            raise errors.ParameterError(f"Path '{path}' is not within base path '{base}'")

        relative_paths.append(str(relative_path))

    return relative_paths


def are_paths_related(a, b):
    """Return True if paths are equal or one is the parent of the other."""
    common_path = os.path.commonpath((a, b))
    absolute_common_path = os.path.abspath(common_path)
    return absolute_common_path == os.path.abspath(a) or absolute_common_path == os.path.abspath(b)



# -*- coding: utf-8 -*-
#
# Copyright 2017-2021 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Test various utilities."""

import os

from renku.core.errors import ParameterError
from renku.core.utils.scm import shorten_message
from renku.core.utils.urls import get_host
from tests.utils import raises


def test_hostname():
    """Test host is set correctly in a different Renku domain."""
    renku_domain = os.environ.get("RENKU_DOMAIN")
    try:
        os.environ["RENKU_DOMAIN"] = "alternative-domain"

        assert "alternative-domain" == get_host(None)
    finally:
        if renku_domain:
            os.environ["RENKU_DOMAIN"] = renku_domain
        else:
            del os.environ["RENKU_DOMAIN"]


def test_shorten_message():
    """Test message is shorten correctly."""

    short_message = "Lorem ipsum dolor sit amet"
    long_message = (
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt "
        "ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco "
        "laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in "
        "voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat "
        "non proident, sunt in culpa qui officia deserunt mollit anim id est laborum."
    )
    max_line = 150

    # The message is modified only when it's longer than the max line length
    assert shorten_message(short_message, max_line) == short_message
    assert shorten_message(long_message, max_line) != long_message
    assert "\n" in shorten_message(long_message, max_line)
    assert shorten_message(long_message, 0) == long_message
    assert shorten_message(long_message, len(long_message)) == long_message
    assert len(shorten_message(long_message, max_line)) > len(long_message)

    # The message is trimmed when it's longer than the max body length
    assert len(shorten_message(long_message, 0, body_length=max_line)) < len(long_message)
    assert len(shorten_message(short_message, 0, body_length=max_line)) == len(short_message)
    assert len(shorten_message(long_message, 0, body_length=0)) == len(long_message)
    assert len(shorten_message(long_message, 0, body_length=len(long_message))) == len(long_message)

    # The the message is wrapped correctly
    assert len(shorten_message(long_message, max_line, body_length=max_line * 2)) < len(long_message)
    assert len(shorten_message(short_message, max_line, body_length=max_line * 2)) == len(short_message)
    assert shorten_message(long_message, max_line, body_length=max_line * 2).count("\n") == 1

    # The message is not distored
    restored_message = shorten_message(long_message, max_line).replace("\n\t", " ")
    assert restored_message == long_message

    restored_message = shorten_message(long_message, max_line, body_length=max_line * 2).replace("\n\t", " ")
    restored_message = restored_message[: max_line * 2 - 3]
    assert long_message.startswith(restored_message)
    assert not restored_message.startswith(long_message)

    # Negative lengths are not accepted
    with raises(ParameterError):
        shorten_message(short_message, -1)
    with raises(ParameterError):
        shorten_message(short_message, max_line, -1)
